@@ -1,22 +1,20 @@
 /* display.c */
 /* Copyright 1995 by Steve Kirkendall */
 
-char id_display[] = "$Id: display.c,v 2.15 1996/04/19 19:13:55 steve Exp $";
+char id_display[] = "$Id: display.c,v 2.17 1997/02/09 22:15:56 steve Exp $";
 
 #include "elvis.h"
 
 /* This is a list of all known display modes.  The last mode is the default. */
 DISPMODE *allmodes[] =
 {
-#ifdef DISPLAY_C
-	&dmc,
-#endif
 #ifdef DISPLAY_HEX
 	&dmhex,
 #endif
 #ifdef DISPLAY_MARKUP
 	&dmhtml,
 	&dmman,
+	&dmtex,
 #endif
 #ifdef DISPLAY_SYNTAX
 	&dmsyntax,
