@@ -1,7 +1,7 @@
 /* exmake.c */
 /* Copyright 1995 by Steve Kirkendall */
 
-char id_make[] = "$Id: exmake.c,v 2.15 1996/05/23 00:07:17 steve Exp $";
+char id_make[] = "$Id: exmake.c,v 2.26 1998/07/31 03:30:01 steve Exp $";
 
 #include "elvis.h"
 
@@ -11,11 +11,13 @@ static RESULT gotoerr(EXINFO *xinf);
 static void errprep(void);
 #endif
 
+static char	*maybedir;	/* directory name extracted from errlist */
 static CHAR	*errfile;	/* name of file where error was detected */
 static long	errline;	/* the line number for an error */
 static CHAR	*errdesc;	/* description of the error */
 static MARK	errnext;	/* used for stepping through the error list */
 
+
 /* This function tries to parse an error message.  Each error message is
  * assumed to fit on a single line.  Within the line, this parser attempts to
  * locate three fields: the source file name, the line number, and the
@@ -23,7 +25,7 @@ static MARK	errnext;	/* used for stepping through the error list */
  *
  * The line is scanned for words.  "Words", for this purpose, are considered
  * to be contiguous strings of non-whitespace characters other than ':'
- * (unless followed by a backslash) or '(' or ')' or any quote character.
+ * (unless followed by a backslash) or any of ( ) , " '
  *
  * If a word happens to be the name of an existing file which is writable
  * by the user, then it is taken to be the filename.  Otherwise, if it contains
@@ -46,21 +48,24 @@ static BOOLEAN parse_errmsg()
 	CHAR		*word;		/* dynamically allocated string */
 	CHAR		*cp;		/* used for scanning line */
 	DIRPERM		perms;		/* permissions of a file */
+	long		top;		/* start of line */
+	WINDOW		errwin;		/* window which displays errors */
 
 	/* if we aren't already reading the errlist, then start at beginning */
 	if (!errnext)
 	{
 		errlist = buffind(toCHAR(ERRLIST_BUF));
 		if (!errlist)
-			return False;
+			goto NoMoreErrors;
 		errnext = markalloc(errlist, 0);
 	}
 
 	/* if we reached the end of the errlist, then say so */
 	if (markoffset(errnext) >= o_bufchars(markbuffer(errnext)))
-		return False;
+		goto NoMoreErrors;
 
 	/* prepare to start scanning */
+	top = markoffset(errnext);
 	if (errfile)
 		safefree(errfile);
 	if (errdesc)
@@ -74,8 +79,7 @@ static BOOLEAN parse_errmsg()
 	for (scanalloc(&cp, errnext); cp && *cp != '\n'; cp && scannext(&cp))
 	{
 		/* is the character legal in a word? */
-		if (!isspace(*cp) && *cp != '(' && *cp != ')' && *cp != ':'
-			&& *cp != '"' && *cp != '\'' && *cp != '`' && *cp != ',')
+		if (!isspace(*cp) && CHARchr("():\"'`,", *cp) == NULL)
 		{
 			/* character in word */
 			buildCHAR(&word, *cp);
@@ -99,6 +103,12 @@ static BOOLEAN parse_errmsg()
 			if (cp)
 				scanprev(&cp);
 		}
+		else if (errfile && errline && (*cp == '"' && *cp == '\'' && *cp != '`'))
+		{
+			/* quotes are allowed at the start of a description */
+			buildCHAR(&word, *cp);
+			continue;
+		}
 
 		/* if we get here, then we aren't in a word.  If no word was
 		 * in progress before this character, then we can ignore this
@@ -108,14 +118,33 @@ static BOOLEAN parse_errmsg()
 			continue;
 
 		/* So we must have a word that we need to process.  Is it the
-		 * name of an existing, writable file?
+		 * name of an existing, writable text file?
 		 */
-		if (!errfile && ((perms = dirperm(tochar8(word))) == DIR_READWRITE
-				|| (o_anyerror && perms == DIR_READONLY)))
+		if (!errfile
+		 && ((perms = dirperm(tochar8(word))) == DIR_READWRITE
+			|| (o_anyerror && perms == DIR_READONLY))
+		 && *ioeol(tochar8(word)) != 'b')
 		{
 			/* this is the name of the source file */
 			errfile = word;
 		}
+		else if (dirperm(tochar8(word)) == DIR_NOTFILE)
+		{
+			/* this may be the name of a directory where
+			 * source files reside */
+			if (maybedir)
+				safefree(maybedir);
+			maybedir = tochar8(word);
+		}
+		else if (!errfile
+		 && maybedir
+		 && ((perms = dirperm(dirpath(maybedir, tochar8(word))) ) == DIR_READWRITE
+			|| (o_anyerror && perms == DIR_READONLY))
+		 && *ioeol(dirpath(maybedir, tochar8(word))) != 'b')
+		{
+			errfile = CHARdup(toCHAR(dirpath(maybedir, tochar8(word))));
+			safefree(word);
+		}
 		else if (!errline && calcnumber(word))
 		{
 			/* this is the line number */
@@ -130,7 +159,7 @@ static BOOLEAN parse_errmsg()
 			safefree(word);
 			word = NULL;
 			while (cp && *cp != '\n' &&
-				(isspace(*cp) || isdigit(*cp) || *cp == ':'))
+				(isspace(*cp) || isdigit(*cp) || *cp == ':' || *cp == '-'))
 			{
 				scannext(&cp);
 			}
@@ -194,9 +223,70 @@ static BOOLEAN parse_errmsg()
 		marksetoffset(errnext, o_bufchars(markbuffer(errnext)));
 	}
 
+	/* If a window is showing the error list, then highlight the current
+	 * line, and leave the cursor at its end.
+	 */
+	errwin = winofbuf(NULL, markbuffer(errnext));
+	if (errwin && !errwin->state->acton)
+	{
+		/* move the cursor to the end of the line */
+		marksetoffset(errwin->cursor, markoffset(errnext) - 2);
+		marksetoffset(errwin->state->top, markoffset(errwin->cursor));
+		marksetoffset(errwin->state->bottom, markoffset(errwin->cursor));
+
+		/* select the line, so it appears highlighted */
+		if (!errwin->seltop)	errwin->seltop = markdup(errwin->cursor);
+		if (!errwin->selbottom)	errwin->selbottom = markdup(errwin->cursor);
+		marksetoffset(errwin->seltop, top);
+		marksetoffset(errwin->selbottom, markoffset(errnext) - 1);
+		errwin->selleft = 0;
+		errwin->selright = INFINITY;
+		errwin->seltype = 'l';
+	}
+
+	/* Set the changepos to end of the message.  This is so that if, after
+	 * finding one or more messages, the user creates a new window for
+	 * viewing the error list, the new window's cursor will start at the
+	 * end of the current message instead of at the top of the buffer.
+	 */
+	markaddoffset(errnext, -1L);
+	bufwilldo(errnext, False);
+	markaddoffset(errnext, 1L);
+
 	/* return True since there was a line for us to parse */
 	scanfree(&cp);
 	return True;
+
+NoMoreErrors:
+	/* if a window is showing the error list, then unhighlight the old
+	 * line (if any was highlighted)
+	 */
+	errlist = buffind(toCHAR(ERRLIST_BUF));
+	errwin = errlist ? winofbuf(NULL, errlist) : NULL;
+	if (errwin && !errwin->state->acton)
+	{
+		/* leave the cursor after the last error message */
+		if (o_bufchars(errlist) > 2)
+			marksetoffset(errwin->cursor, o_bufchars(errlist) - 2);
+		else
+			marksetoffset(errwin->cursor, 0L);
+		marksetoffset(errwin->state->top, markoffset(errwin->cursor));
+		marksetoffset(errwin->state->bottom, markoffset(errwin->cursor));
+
+		/* unselect the line, so nothing appears highlighted */
+		if (errwin->seltop)
+		{
+			assert(errwin->selbottom);
+			markfree(errwin->seltop);
+			markfree(errwin->selbottom);
+			errwin->seltop = errwin->selbottom = NULL;
+		}
+
+		/* make sure the window will be redrawn */
+		errwin->di->logic = DRAW_CHANGED;
+	}
+
+	return False;
 }
 
 
@@ -209,6 +299,14 @@ static RESULT gotoerr(xinf)
 {
 	BUFFER	blamebuf;
 	long	blameline;
+	WINDOW	errwin;
+
+	/* if there is a window showing the error buffer, then search forward
+	 * from its cursor. (Else search forward from end of previous error.)
+	 */
+	errwin = (errnext ? winofbuf(NULL, markbuffer(errnext)) : NULL);
+	if (errwin)
+		marksetoffset(errnext, markoffset(errwin->cursor));
 
 	/* parse lines from the errlist buffer until we find an error message */
 	do
@@ -265,10 +363,9 @@ static void errprep()
 	MARKBUF	top;
 	MARKBUF	end;
 
-	/* create the "errlist" buffer */
-	buf = bufalloc(toCHAR(ERRLIST_BUF), 0);
-	o_internal(buf) = True;
-	o_undolevels(buf) = 0;
+	/* find the "Elvis error list" buffer */
+	buf = buffind(toCHAR(ERRLIST_BUF));
+	assert(buf != NULL);
 
 	/* if it has old text in it, discard that text */
 	if (o_bufchars(buf) > 0)
@@ -285,6 +382,13 @@ static void errprep()
 		errnext = NULL;
 	}
 
+	/* clobber the maybedir */
+	if (maybedir)
+	{
+		safefree(maybedir);
+		maybedir = NULL;
+	}
+
 	/* remember how many lines each buffer has now */
 	for (buf = NULL; (buf = buflist(buf)) != NULL; )
 	{
@@ -296,16 +400,37 @@ static void errprep()
 RESULT	ex_errlist(xinf)
 	EXINFO	*xinf;
 {
+	BUFFER	errbuf;
+	WINDOW	errwin;
+	MARKBUF	from;
+	BOOLEAN	retval;
+
 	assert(xinf->command == EX_ERRLIST);
 
 	/* was a filename given? */
-	if (xinf->nfiles > 0)
+	if (xinf->nfiles > 0 || xinf->rhs)
 	{
 		/* prepare to create a new error list */
 		errprep();
 
 		/* load the buffer from the named file */
-		(void)bufload(toCHAR(ERRLIST_BUF), xinf->file[0], True);
+		errbuf = bufalloc(toCHAR(ERRLIST_BUF), 0, True);
+		retval = bufread(marktmp(from, errbuf, 0), xinf->rhs ? tochar8(xinf->rhs) : xinf->file[0]);
+
+		/* turn off the "modified" flag on the (Elvis error list) buf */
+		o_modified(errbuf) = False;
+
+		/* if there are any windows showing this buffer, move its cursor
+		 * to offset 0
+		 */
+		for (errwin = NULL; (errwin = winofbuf(errwin, errbuf)) != NULL; )
+		{
+			marksetoffset(errwin->cursor, 0L);
+		}
+
+		/* if failed to read errors, then return RESULT_ERROR */
+		if (!retval)
+			return RESULT_ERROR;
 	}
 
 	/* move the cursor to the next error */
@@ -322,7 +447,9 @@ RESULT	ex_make(xinf)
 	int	nio;	/* number of characters read */
 	MARKBUF	start, end;/* ends of errlist buffer, used for appending */
 	BOOLEAN	origrefresh;
+	long	origpollfreq;
 	BUFFER	buf;
+	WINDOW	errwin;	/* window which is displaying errors */
 
 	assert(xinf->command == EX_MAKE || xinf->command == EX_CC);
 
@@ -354,6 +481,11 @@ RESULT	ex_make(xinf)
 	}
 
 	/* output the command name, so the user knows what's happening */
+	origrefresh = o_exrefresh;
+	o_exrefresh = True;
+	origpollfreq = o_pollfrequency;
+	o_pollfrequency = 1;
+	(void)guipoll(True);
 	drawextext(xinf->window, str, (int)CHARlen(str));
 	drawextext(xinf->window, toCHAR("\n"), 1);
 
@@ -366,6 +498,8 @@ RESULT	ex_make(xinf)
 	if (!prgopen(tochar8(str), False, True) || !prggo())
 	{
 		/* failed -- error message already given */
+		o_exrefresh = origrefresh;
+		o_pollfrequency = origpollfreq;
 		return RESULT_ERROR;
 	}
 	buf = buffind(toCHAR(ERRLIST_BUF));
@@ -376,9 +510,7 @@ RESULT	ex_make(xinf)
 	}
 	io = (CHAR *)safealloc(1024, sizeof(CHAR));
 	(void)marktmp(end, buf, 0);
-	origrefresh = o_exrefresh;
-	o_exrefresh = True;
-	while ((nio = prgread(io, 1024)) > 0)
+	while (!guipoll(False) && (nio = prgread(io, 1024)) > 0)
 	{
 		/* show it on the screen */
 		drawextext(xinf->window, io, nio);
@@ -387,9 +519,21 @@ RESULT	ex_make(xinf)
 		bufreplace(&end, &end, io, nio);
 		markaddoffset(&end, nio);
 	}
-	o_exrefresh = origrefresh;
 	(void)prgclose();
 	safefree(io);
+	o_exrefresh = origrefresh;
+	o_pollfrequency = origpollfreq;
+
+	/* turn off the "modified" flag on the (Elvis error list) buf */
+	o_modified(buf) = False;
+
+	/* if there is a window showing this buffer, move its cursor
+	 * to offset 0
+	 */
+	for (errwin = NULL; (errwin = winofbuf(errwin, buf)) != NULL; )
+	{
+		marksetoffset(errwin->cursor, 0L);
+	}
 
 	/* move the cursor to the first error */
 	return gotoerr(xinf);
