@@ -1,7 +1,7 @@
 /* gui.c */
 /* Copyright 1995 by Steve Kirkendall */
 
-char id_gui[] = "$Id: gui.c,v 2.8 1996/05/23 00:10:56 steve Exp $";
+char id_gui[] = "$Id: gui.c,v 2.13 1997/12/24 03:12:52 steve Exp $";
 
 #include "elvis.h"
 
@@ -17,15 +17,10 @@ void guimoveto(win, column, row)
 	int	column;	/* column where cursor should be placed */
 	int	row;	/* row where cursor should be placed */
 {
-	/* perform the moveto, unless the cursor is already there */
-#if 0
-	if (column != win->cursx || row != win->cursy)
-#endif
-	{
-		(*gui->moveto)(win->gw, column, row);
-		win->cursx = column;
-		win->cursy = row;
-	}
+	/* perform the moveto */
+	(*gui->moveto)(win->gw, column, row);
+	win->cursx = column;
+	win->cursy = row;
 }
 
 /* This function calls the GUI's draw() function, and then updates elvis'
@@ -164,7 +159,10 @@ BOOLEAN guipoll(reset)
 	if (reset)
 	{
 		pollctr = 0;
-		return False;
+		if (gui && gui->poll)
+			return (*gui->poll)(reset);
+		else
+			return False;
 	}
 
 	/* if there is no poll() function, or pollfrequency indicates that
@@ -180,10 +178,27 @@ BOOLEAN guipoll(reset)
 	pollctr = 0;
 
 	/* call the GUI's poll() function, and return its value */
-	reset = (*gui->poll)();
+	reset = (*gui->poll)(reset);
 	if (reset)
 	{
 		msg(MSG_ERROR, "aborted");
 	}
 	return reset;
 }
+
+/* ring the bell, if there is one.  Limit it to one ding per eventdraw() */
+void guibeep(win)
+	WINDOW	win;	/* window to ding, or NULL to indicate an eventdraw() */
+{
+	static BOOLEAN	dingable = True;
+
+	if (!win)
+	{
+		dingable = True;
+	}
+	else if (gui && gui->beep && dingable)
+	{
+		(*gui->beep)(win->gw);
+		dingable = False;
+	}
+}
