@@ -1,7 +1,7 @@
 /* input.c */
 /* Copyright 1995 by Steve Kirkendall */
 
-char id_input[] = "$Id: input.c,v 2.35 1996/05/27 07:58:16 steve Exp $";
+char id_input[] = "$Id: input.c,v 2.62 1998/11/25 19:20:58 steve Exp $";
 
 #include "elvis.h"
 
@@ -45,6 +45,7 @@ typedef struct
 #if USE_PROTOTYPES
 static void cleanup(WINDOW win, BOOLEAN del, BOOLEAN backsp, BOOLEAN yank);
 static void addchar(MARK cursor, MARK top, MARK bottom, INPUTINFO *info);
+static BOOLEAN tryabbr(WINDOW win, _CHAR_ nextc);
 static RESULT perform(WINDOW win);
 static RESULT parse(_CHAR_ key, void *info);
 static ELVCURSOR shape(WINDOW win);
@@ -64,16 +65,25 @@ static void cleanup(win, del, backsp, yank)
 	BOOLEAN	yank;	/* if True, yank input text into ". buffer */
 {
 	/* delete the excess in the edited region */
-	if (del && markoffset(win->state->cursor) < markoffset(win->state->bottom))
+	if (del && markoffset(win->state->cursor) < markoffset(win->state->bottom)
+	        && markoffset(win->state->top) < markoffset(win->state->bottom))
 	{
 		bufreplace(win->state->cursor, win->state->bottom, NULL, 0);
 	}
 	else
 	{
 		marksetoffset(win->state->bottom, markoffset(win->state->cursor));
+		if (markoffset(win->state->top) > markoffset(win->state->cursor))
+			marksetoffset(win->state->top, markoffset(win->state->cursor));
 	}
 	assert(markoffset(win->state->cursor) == markoffset(win->state->bottom));
 
+	/* save the newly input text in the "previous input" buffer */
+	if (yank)
+	{
+		cutyank('.', win->state->top, win->state->bottom, 'c', False);
+	}
+
 	/* move the cursor back one character, unless it is already at the
 	 * start of a line.
 	 */
@@ -84,12 +94,8 @@ static void cleanup(win, del, backsp, yank)
 		{
 			markaddoffset(win->state->cursor, 1);
 		}
-	}
-
-	/* save the newly input text in the "previous input" buffer */
-	if (yank)
-	{
-		cutyank('.', win->state->top, win->state->bottom, 'c', False);
+		marksetoffset(win->state->top, markoffset(win->state->cursor));
+		marksetoffset(win->state->bottom, markoffset(win->state->cursor));
 	}
 
 	/* Force the screen to be regenerated */
@@ -133,6 +139,106 @@ static void addchar(cursor, top, bottom, info)
 	}
 }
 
+
+/* This function attempts to expand an abbreviation at the cursor location,
+ * if there is one.  If so, it deletes the short form, and pushes the long
+ * form an following character back onto the type-ahead buffer.  Else it
+ * returns False.
+ */
+static BOOLEAN tryabbr(win, nextc)
+	WINDOW	win;	/* window where abbreviation may need expanding */
+	_CHAR_	nextc;	/* character after the word */
+{
+	MARKBUF	from, to;
+	CHAR	*cp, *build;
+	long	slen, llen, tlen;
+	CHAR	cbuf[1];
+
+
+#if 0 /* this would prevent abbrs from being expanded on command lines */
+	if((win->state->flags & (ELVIS_POP|ELVIS_ONCE|ELVIS_1LINE)) ==
+		 (ELVIS_POP|ELVIS_ONCE|ELVIS_1LINE)) { /* nishi */
+	    return False;
+	}
+#endif
+
+	/* Try to do an abbreviation.  To do this, we collect
+	 * characters backward to the preceding whitespace.  We
+	 * go to the preceding whitespace because abbreviations
+	 * can't contain whitespace; we know we'll never need
+	 * more characters to recognize an abbreviation.  We
+	 * collect the characters backwards just because it is
+	 * easier.
+	 */
+	for (scanalloc(&cp, win->state->cursor), build = NULL;
+	     cp && scanprev(&cp) && !isspace(*cp);
+	     )
+	{
+		buildCHAR(&build, *cp);
+	}
+	scanfree(&cp);
+	if (build)
+	{
+		cp = mapabbr(build, &slen, &llen, (BOOLEAN)(win->state->acton != NULL));
+		if (cp)
+		{
+			/* yes, we have an abbreviation! */
+
+			/* count the characters in the long form which appear
+			 * before the first control character.
+			 */
+			for (tlen = 0; tlen < llen && cp[tlen] >= ' '; tlen++)
+			{
+			}
+
+			/* replace the short form with the plain text part of
+			 * the long form.
+			 */
+			cleanup(win, True, False, False);
+			(void)marktmp(from, markbuffer(win->state->cursor), markoffset(win->state->cursor) - slen);
+			(void)marktmp(to, markbuffer(win->state->cursor), markoffset(win->state->cursor));
+			bufreplace(&from, &to, cp, tlen);
+			cp += tlen;
+			llen -= tlen;
+			markaddoffset(&from, tlen);
+
+			/* handle the final character (space or tab typed by
+			 * the user) and any remaining characters from the
+			 * long form of the abbreviation.
+			 */
+			if (nextc)
+			{
+				cbuf[0] = nextc;
+				if (llen > 0)
+				{
+					/* stuff the long form, and the user's
+					 * non-alnum character, into the queue
+					 */
+					mapunget(cbuf, 1, False);
+					mapunget(cp, (int)llen, False);
+				}
+				else
+				{
+					/* insert the final character into the
+					 * text
+					 */
+					bufreplace(&from, &from, cbuf, 1L);
+					markaddoffset(&from, 1L);
+				}
+			}
+
+			/* move cursor to where word goes */
+			marksetoffset(win->state->bottom, markoffset(&from));
+			marksetoffset(win->state->cursor, markoffset(&from));
+			safefree(build);
+			return True;
+		}
+		safefree(build);
+	}
+	return False;
+}
+
+
 /* This function performs an input-mode command.  Usually, this will be a
  * character to insert/replace in the text.
  */
@@ -144,16 +250,19 @@ static RESULT perform(win)
 	MARK	  cursor = state->cursor;
 	MARK	  tmp;
 	MARKBUF	  from, to;
-	CHAR	  *cp, *build;
-	char	  *littlecp, ch;
+	CHAR	  *cp;
 	EXINFO	  xinfb;
 	VIINFO	  vinfb;
+	long	  col, len;
+#ifdef FEATURE_COMPLETE
+	CHAR	  *end;
 	union
 	{
 		char	  partial[256];
 		CHAR	  full[256];
 	}	  name;
-	long	  col, len;
+	char	  *littlecp, ch, peek;
+#endif
 
 	safeinspect();
 
@@ -191,51 +300,14 @@ static RESULT perform(win)
 		 * (Note: Since we never expand abbreviations except in the
 		 * main buffer or the ex history buffer, we can skip it if
 		 * we're editing some other buffer such as regexp history.
-		 * Assume only the ex history buffer has inputtab=filename.)
+		 * Assume only the ex history buffer has inputtab=ex.)
 		 */
 		if (!isalnum(ii->arg)
 		 && ii->arg != '_'
-		 && (state->acton == NULL || o_inputtab(markbuffer(cursor)) == 'f'))
+		 && (state->acton == NULL || o_inputtab(markbuffer(cursor)) == 'e'))
 		{
-			/* Try to do an abbreviation.  To do this, we collect
-			 * characters backward to the preceding whitespace.  We
-			 * go to the preceding whitespace because abbreviations
-			 * can't contain whitespace; we know we'll never need
-			 * more characters to recognize an abbreviation.  We
-			 * collect the characters backwards just because it is
-			 * easier.
-			 */
-			for (scanalloc(&cp, cursor), build = NULL;
-			     cp && scanprev(&cp) && !isspace(*cp);
-			     )
-			{
-				buildCHAR(&build, *cp);
-			}
-			scanfree(&cp);
-			if (build)
-			{
-				cp = mapabbr(build, &col, &len, (BOOLEAN)(state->acton != NULL));
-				if (cp)
-				{
-					cleanup(win, True, False, False);
-					(void)marktmp(from, markbuffer(cursor), markoffset(cursor) - col);
-					(void)marktmp(to, markbuffer(cursor), markoffset(cursor));
-#if 0
-					bufreplace(&from, &to, cp, len);
-					marksetoffset(state->bottom, markoffset(&from) + len);
-					marksetoffset(cursor, markoffset(&from) + len);
-#else
-					bufreplace(&from, &to, NULL, 0);
-					mapunget(&ii->arg, 1, False);
-					mapunget(cp, (int)len, False);
-					marksetoffset(state->bottom, markoffset(&from));
-					marksetoffset(cursor, markoffset(&from));
-					safefree(build);
-					break;
-#endif
-				}
-				safefree(build);
-			}
+			if (tryabbr(win, ii->arg))
+				break;
 		}
 		/* fall through... */
 
@@ -251,12 +323,12 @@ static RESULT perform(win)
 		if (!isspace(ii->arg)
 		 && cursor == win->cursor
 		 && win->md->wordwrap
-		 && o_linelength(markbuffer(cursor)) > 0
-		 && dispmark2col(win) >= o_linelength(markbuffer(cursor)))
+		 && o_textwidth(markbuffer(cursor)) > 0
+		 && dispmark2col(win) >= o_textwidth(markbuffer(cursor)))
 		{
 			/* Figure out where the current word started */
 			for (scanalloc(&cp, cursor);
-			     scanprev(&cp) && !isspace(*cp);
+			     cp && scanprev(&cp) && !isspace(*cp);
 			     )
 			{
 			}
@@ -265,6 +337,10 @@ static RESULT perform(win)
 				to = *scanmark(&cp);
 				markaddoffset(&to, 1L);
 			}
+			else
+			{
+				to = *dispmove(win, 0L, 0L);
+			}
 
 			/* Locate the front of this line.  We won't look past
 			 * back before this character.
@@ -324,101 +400,180 @@ static RESULT perform(win)
 
 	  case INP_NEWLINE:
 		cleanup(win, True, False, False);
-		ii->arg = '\n';
-		ii->cmd = INP_NORMAL;
-		perform(win);
-		dispindent(win, cursor, -1);
+		if (!tryabbr(win, '\n'))
+		{
+			ii->arg = '\n';
+			ii->cmd = INP_NORMAL;
+			perform(win);
+			dispindent(win, cursor, -1);
+		}
 		ii->backsp = ii->prev = '\0';
 		break;
 
 	  case INP_TAB:
-		switch (o_inputtab(markbuffer(cursor)))
+		if (!tryabbr(win, '\t'))
 		{
-		  case 't':
-			/* insert a tab */
-			addchar(cursor, state->top, state->bottom, ii);
-			break;
-
-		  case 's':
-			/* insert enough spaces to look like a tab */
-			col = dispmark2col(win);
-			ii->arg = ' ';
-			do
+			switch (o_inputtab(markbuffer(cursor)))
 			{
+#ifndef FEATURE_COMPLETE
+			  case 'e':
+			  case 'f':
+			  case 'i':
+#endif /* not FEATURE_COMPLETE */
+			  case 't':
+				/* insert a tab */
 				addchar(cursor, state->top, state->bottom, ii);
-			} while ((++col) % o_tabstop(markbuffer(cursor)) != 0);
-			break;
-
-		  case 'f':
-			/* filename completion */
-
-			/* if at start of input, then fail */
-			if (markoffset(cursor) == markoffset(state->top))
-			{
-				guibeep(win);
 				break;
-			}
 
-			/* locate the previous character */
-			tmp = markdup(cursor);
-			markaddoffset(tmp, -1);
+			  case 's':
+				/* insert enough spaces to look like a tab */
+				col = dispmark2col(win);
+				ii->arg = ' ';
+				do
+				{
+					addchar(cursor, state->top, state->bottom, ii);
+				} while ((++col) % o_tabstop(markbuffer(cursor)) != 0);
+				break;
+#ifdef FEATURE_COMPLETE
+			  case 'i':
+				/* try to find a matching tag name */
+				cp = tagcomplete(win, cursor);
 
-			/* if previous is whitespace, then fail */
-			if (isspace(scanchar(tmp)) || scanchar(tmp) == '(')
-			{
+				/* If no matching tag, or already at end of
+				 * complete tag, then insert a literal tab
+				 * character, not a space.
+				 */
+				if (*cp == ' ')
+					*cp = '\t';
+
+				/* For other values, delete the trailing
+				 * whitespace which would normally be appended.
+				 */
+				end = CHARchr(cp, ' ');
+				if (end)
+					*end = '\0';
+
+				/* Now... if we have anything to add, add it */
+				if (*cp)
+				{
+					len = CHARlen(cp);
+					col = markoffset(cursor) + len;
+					bufreplace(cursor, win->state->bottom, cp, len);
+					marksetoffset(cursor, col);
+					marksetoffset(win->state->bottom, col);
+				}
+				break;
+
+			  case 'e':
+				/* Assume this is an ex command line, and try
+				 * to do something useful.
+				 */
+				tmp = markdup(cursor);
+				(void)marksetline(tmp, markline(cursor));
+				cp = excomplete(win, tmp, cursor);
 				markfree(tmp);
-				guibeep(win);
-				break;
-			}
+				if (cp)
+				{
+					/* excomplete() found something useful,
+					 * or returned "\t" to make the <Tab>
+					 * be a literal tab character.  Copy
+					 * it into the buffer.
+					 */
+					len = CHARlen(cp);
+					col = markoffset(cursor) + len;
+					bufreplace(cursor, win->state->bottom, cp, len);
+					marksetoffset(cursor, col);
+					marksetoffset(win->state->bottom, col);
+					break;
+				}
+				/* else fall through to filename completion */
 
-			/* collect the partial name into char array */
-			littlecp = &name.partial[QTY(name.partial)];
-			*--littlecp = '\0';
-			ch = scanchar(tmp);
-			do
-			{
-				*--littlecp = ch;
+			  case 'f':
+				/* filename completion */
+
+				/* if at start of input, then fail */
+				if (markoffset(cursor) == markoffset(state->top))
+				{
+					guibeep(win);
+					break;
+				}
+
+				/* locate the previous character */
+				tmp = markdup(cursor);
 				markaddoffset(tmp, -1);
-				ch = (markoffset(tmp) >= markoffset(state->top)) ? scanchar(tmp) : ' ';
-			} while (!isspace(ch) && ch != '(');
-			markaddoffset(tmp, 1);
 
-			/* try to expand the filename */
-			littlecp = iofilename(littlecp, (ch == '(') ? ')' : '\t');
-			if (!littlecp)
-			{
+				/* if previous char can't be in name, fail */
+				if (CHARchr(toCHAR(" \t\n()<>"), scanchar(tmp)))
+				{
+					markfree(tmp);
+					guibeep(win);
+					break;
+				}
+
+				/* collect the partial name into char array */
+				littlecp = &name.partial[QTY(name.partial)];
+				*--littlecp = '\0';
+				ch = scanchar(tmp);
+				do
+				{
+					*--littlecp = ch;
+					markaddoffset(tmp, -1L);
+					ch = (markoffset(tmp) >= markoffset(state->top)) ? scanchar(tmp) : '\n';
+					if (markoffset(tmp) - 1 >= markoffset(state->top))
+					{
+						from = *tmp;
+						markaddoffset(&from, -1L);
+						peek = scanchar(&from);
+					}
+					else
+						peek = 'x'; /* not backslash */
+				} while (ch != '\n' && ch && ch != '\t' &&
+				   (!(ch == ' ' || ch == '<' || ch == '>' || ch == '(' || ch == ')')
+					|| peek == '\\'));
+				markaddoffset(tmp, 1);
+
+				/* try to expand the filename */
+				littlecp = iofilename(littlecp,
+					(ch == '(') ? ')' :
+					o_inputtab(markbuffer(cursor)) == 'e' ? ' ' :
+					'\t');
+				if (!littlecp)
+				{
+					markfree(tmp);
+					guibeep(win);
+					break;
+				}
+
+				/* name found -- replace old word with expanded name */
+				for (cp = name.full, col = 0;
+				     (*cp++ = *littlecp++) != '\0'; /* yes, ASSIGNMENT! */
+				     col++)
+				{
+				}
+				bufreplace(tmp, win->state->bottom, name.full, col);
+				marksetoffset(cursor, markoffset(tmp) + col);
+				marksetoffset(win->state->bottom, markoffset(cursor));
 				markfree(tmp);
-				guibeep(win);
-				break;
+#endif /* FEATURE_COMPLETE */
 			}
-
-			/* name found -- replace old word with expanded name */
-			for (cp = name.full, col = 0;
-			     (*cp++ = *littlecp++) != '\0'; /* yes, ASSIGNMENT! */
-			     col++)
-			{
-			}
-			bufreplace(tmp, cursor, name.full, col);
-			marksetoffset(cursor, markoffset(tmp) + col);
-			markfree(tmp);
 		}
 		ii->backsp = ii->prev = '\0';
 		break;
 
 	  case INP_ONEVI:
-		cleanup(win, False, False, True);
+		cleanup(win, True, False, True);
 		vipush(win, ELVIS_ONCE, NULL);
 		ii->backsp = ii->prev = '\0';
 		ii->setbottom = True;
 		break;
 
 	  case INP_MULTIVI:
+		(void)tryabbr(win, '\0');
 		cleanup(win, True, True, True);
 		win->state->flags |= ELVIS_POP;
 		ii->backsp = ii->prev = '\0';
 
-#if 1
-		/* if blank line in autoindent mode, then delete any whitespace */
+		/* if blank line in autoindent mode, then delete whitespace */
 		if (o_autoindent(markbuffer(cursor)))
 		{
 			for (from = *dispmove(win,0L,0L), scanalloc(&cp, &from);
@@ -426,7 +581,7 @@ static RESULT perform(win)
 			     scannext(&cp))
 			{
 			}
-			if (cp)
+			if (cp && *cp == '\n')
 			{
 				to = *scanmark(&cp);
 				scanfree(&cp);
@@ -441,7 +596,6 @@ static RESULT perform(win)
 				scanfree(&cp);
 			}
 		}
-#endif
 		break;
 
 	  case INP_BACKSPACE:
@@ -551,8 +705,12 @@ static RESULT perform(win)
 		xinfb.bang = True;
 
 		/* execute the command */
+		len = o_bufchars(markbuffer(cursor)) - markoffset(cursor);
+		assert(len >= 0);
 		(void)ex_shift(&xinfb);
 		ii->backsp = ii->prev = '\0';
+		assert(o_bufchars(markbuffer(cursor)) >= len);
+		marksetoffset(cursor, o_bufchars(markbuffer(cursor)) - len);
 		break;
 
 	  case INP_EXPOSE:
@@ -586,6 +744,9 @@ static RESULT perform(win)
 
 	}
 
+	/* set wantcol to the cursor's current column */
+	win->wantcol = dispmark2col(win);
+
 	/* prepare for next command */
 	ii->cmd = INP_NORMAL;
 
@@ -601,6 +762,15 @@ static RESULT parse(key, info)
 	void	*info;	/* current parsing state */
 {
 	INPUTINFO *ii = (INPUTINFO *)info;
+	CHAR	sym;	/* visible symbol, e.g. '^' after a ^V */
+
+	/* maybe adjust the edit bounds */
+	if (ii->setbottom)
+	{
+		marksetoffset(windefault->state->bottom,
+			markoffset(windefault->state->cursor));
+		ii->setbottom = False;
+	}
 
 	/* parse the input command */
 	if (ii->cmd == INP_HEX1 || ii->cmd == INP_HEX2)
@@ -661,6 +831,7 @@ static RESULT parse(key, info)
 	else
 	{
 		ii->arg = key;
+		sym = '\0';
 		switch (key)
 		{
 		  case ELVCTRL('@'):	ii->cmd = INP_PREVIOUS;		break;
@@ -670,24 +841,31 @@ static RESULT parse(key, info)
 		  case ELVCTRL('H'):	ii->cmd = INP_BACKSPACE;	break;
 		  case ELVCTRL('I'):	ii->cmd = INP_TAB;		break;
 		  case ELVCTRL('J'):	ii->cmd = INP_NEWLINE;		break;
-		  case ELVCTRL('K'):	ii->cmd = INP_DIG1;		break;
-		  case ELVCTRL('L'):	ii->cmd = INP_EXPOSE;		break;
+		  case ELVCTRL('K'):	ii->cmd = INP_DIG1, sym = '+';	break;
 		  case ELVCTRL('M'):	ii->cmd = INP_NEWLINE;		break;
 		  case ELVCTRL('O'):	ii->cmd = INP_ONEVI;		break;
 		  case ELVCTRL('P'):	ii->cmd = INP_PUT;		break;
 		  case ELVCTRL('R'):	ii->cmd = INP_EXPOSE;		break;
 		  case ELVCTRL('T'):	ii->cmd = INP_SHIFTR;		break;
 		  case ELVCTRL('U'):	ii->cmd = INP_BACKLINE;		break;
-		  case ELVCTRL('V'):	ii->cmd = INP_QUOTE;		break;
+		  case ELVCTRL('V'):	ii->cmd = INP_QUOTE, sym = '^';	break;
 		  case ELVCTRL('W'):	ii->cmd = INP_BACKWORD;		break;
-		  case ELVCTRL('X'):	ii->cmd = INP_HEX1;		break;
+		  case ELVCTRL('X'):	ii->cmd = INP_HEX1, sym = '$';	break;
 		  case ELVCTRL('['):	ii->cmd = INP_MULTIVI;		break;
 		  default:		ii->cmd = INP_NORMAL;
 		}
 
 		/* ^V, ^X, and ^K require more keystrokes... */
-		if (ii->cmd == INP_QUOTE || ii->cmd == INP_HEX1 || ii->cmd == INP_DIG1)
+		if (sym)
 		{
+			/* display a symbol on the screen */
+			ii->arg = sym;
+			addchar(windefault->state->cursor,
+				windefault->state->top,
+				windefault->state->bottom, ii);
+			markaddoffset(windefault->state->cursor, -1);
+
+			/* disable maps during extra character[s] */
 			windefault->state->mapflags |= MAP_DISABLE;
 			return RESULT_MORE;
 		}
@@ -751,6 +929,7 @@ void inputpush(win, flags, mode)
 	_char_		mode;	/* 'R' for replace, or anything else for insert */
 {
 	/* push the state */
+	flags |= ELVIS_REGION;
 	statepush(win, flags);
 
 	/* initialize the state */
@@ -826,6 +1005,7 @@ void inputchange(win, from, to, linemd)
 {
 	MARKBUF	tmp;
 	CHAR	ch;
+	BOOLEAN	ctrl_o = False;	/* JohnW 19/06/96 */
 
 	assert(markbuffer(from) == markbuffer(win->state->cursor));
 	assert(markbuffer(from) == markbuffer(to));
@@ -838,6 +1018,10 @@ void inputchange(win, from, to, linemd)
 	{
 		inputpush(win, 0, 'i');
 	}
+	else
+	{
+		ctrl_o = True;
+	}
 
 	/* replace the last char with '$', if there is a last char
 	 * and it is on the same line.  If it is on a different line,
@@ -847,7 +1031,8 @@ void inputchange(win, from, to, linemd)
 	{
 		/* do nothing */
 	}
-	else if (markoffset(to) > markoffset(dispmove(win, 0, INFINITY)))
+	else if (ctrl_o || /* JohnW 19/06/96 */
+		markoffset(to) > markoffset(dispmove(win, 0, INFINITY)))
 	{
 		/* delete the old text */
 		if (linemd)
@@ -902,6 +1087,9 @@ void inputbeforeenter(win)
 	/* Delete stuff from after the cursor */
 	cleanup(win, True, False, False);
 
+	/* Attempt to expand an abbreviation at the end of the line */
+	(void)tryabbr(win, '\0');
+
 	/* adjust the endpoints of the edited area to be whole lines */
 	marksetoffset(win->state->top,
 		markoffset((*dmnormal.move)(win, win->state->top, 0, 0, False)));
