@@ -1,7 +1,7 @@
 /* lowbuf.c */
 /* Copyright 1995 by Steve Kirkendall */
 
-char id_lowbuf[] = "$Id: lowbuf.c,v 2.17 1996/05/23 00:19:50 steve Exp $";
+char id_lowbuf[] = "$Id: lowbuf.c,v 2.31 1998/11/25 01:28:06 steve Exp $";
 
 /* This file contains functions which divide the session file into several
  * buffers.  These functions use session.c, and are used by buffer.c
@@ -9,6 +9,11 @@ char id_lowbuf[] = "$Id: lowbuf.c,v 2.17 1996/05/23 00:19:50 steve Exp $";
 
 #include "elvis.h"
 
+/* These control the sizes of some tiny caches for storing line offsets and
+ * info about random offsets.  You can disable these by making them 0.  This
+ * is intended to make lowline() run faster.
+ */
+#define LINECACHE	4
 
 #if USE_PROTOTYPES
 static short checksum(BLK *blk);
@@ -19,6 +24,10 @@ static void unlockchars(_BLKNO_ bufinfo, _LBLKNO_ lblkno, _BLKNO_ blkno, int chg
 static void helpinit(_BLKNO_ bufinfo, void (*bufproc)(_BLKNO_ bufinfo, long nchars, long nlines, long changes, long prevloc, CHAR *name));
 #endif
 
+#if LINECACHE
+static void clobbercache P_((_BLKNO_ dst));
+#endif
+
 /******************************************************************************/
 /* Some internal functions...						      */
 
@@ -39,7 +48,15 @@ static short checksum(blk)
 	     scan != blk->sumshorts;
 	     )
 	{
-		sum += *--scan;
+		/* We know that o_blksize is a power of 2, >= 256.  We can
+		 * safely assume that sizeof(short) is also a power of 2, <=16,
+		 * so we can safely unroll this loop.
+		 */
+		sum += scan[-1] + scan[-2] + scan[-3] + scan[-4]
+			+ scan[-5] + scan[-6] + scan[-7] + scan[-8]
+			+ scan[-9] + scan[-10] + scan[-11] + scan[-12]
+			+ scan[-13] + scan[-14] + scan[-15] + scan[-16];
+		scan -= 16;
 	}
 
 	/* restore the old sum, but then return the new sum */
@@ -166,7 +183,7 @@ static BLKNO insblock(blklist, before, chars, nchars, nlines)
 		assert(before == 0);
 
 		/* give the buffer its first blkno */
-		blklist = sesalloc(0);
+		blklist = sesalloc(0, SES_BLKLIST);
 
 		/* make the new CHARS block be its first entry */
 		(void)seslock(blklist, True, SES_BLKLIST);
@@ -350,7 +367,7 @@ static void helpinit(bufinfo, bufproc)
 	int	i;
 
 	/* mark the bufinfo block as being "allocated", and lock it in memory */
-	next = sesalloc(bufinfo);
+	next = sesalloc(bufinfo, SES_BUFINFO);
 	assert(next == bufinfo);
 	(void)seslock(bufinfo, False, SES_BUFINFO);
 	binfo = sesblk(bufinfo);
@@ -361,7 +378,7 @@ static void helpinit(bufinfo, bufproc)
 	for (blklist = binfo->bufinfo.first; blklist; blklist = next)
 	{
 		/* mark the blklist block as being "allocated", & lock it */
-		(void)seslock(sesalloc(blklist), False, SES_BLKLIST);
+		(void)seslock(sesalloc(blklist, SES_BLKLIST), False, SES_BLKLIST);
 		blist = sesblk(blklist);
 
 		/* For each chars block mentioned in the BLKLIST... */
@@ -371,7 +388,7 @@ static void helpinit(bufinfo, bufproc)
 			assert(blist->blklist.blk[i].nlines <= blist->blklist.blk[i].nchars);
 
 			/* mark the CHARS block as being "allocated" */
-			next = sesalloc(blist->blklist.blk[i].blkno);
+			next = sesalloc(blist->blklist.blk[i].blkno, SES_CHARS);
 			assert(next == blist->blklist.blk[i].blkno);
 
 			/* count lines & chars in block */
@@ -430,7 +447,7 @@ void lowinit(bufproc)
 	while (blkno)
 	{
 		/* mark the block as allocated, and lock it into memory */
-		(void)seslock(sesalloc(blkno), False, SES_SUPER2);
+		(void)seslock(sesalloc(blkno, SES_SUPER2), False, SES_SUPER2);
 
 		/* for each buffer mentioned in the superblock... */
 		for (i = 0; i < SES_MAXSUPER2; i++)
@@ -467,7 +484,7 @@ BLKNO lowalloc(name)
 	safeinspect();
 
 	/* Allocate a block, and lock it for writing */
-	bufinfo = seslock(sesalloc(0), True, SES_BUFINFO);
+	bufinfo = seslock(sesalloc(0, SES_BUFINFO), True, SES_BUFINFO);
 
 	/* Fill in the block's info */
 	blk = sesblk(bufinfo);
@@ -500,7 +517,7 @@ BLKNO lowalloc(name)
 				/* need to allocate a new super2 */
 				next = seslock(super, True, SES_SUPER);
 				assert(next == super);
-				next = sesalloc(0);
+				next = sesalloc(0, SES_SUPER2);
 				if (super == 0)
 				{
 					blk->super.buf[i] = next;
@@ -543,6 +560,10 @@ BLKNO lowdup(originfo)
 	BLK	*blk, *dupblk = NULL, *dupbiblk;
 	int	i;
 
+#if LINECACHE
+	clobbercache(originfo);
+#endif
+
 	/* Lock the originfo block */
 	(void)seslock(originfo, False, SES_BUFINFO);
 	blk = sesblk(originfo);
@@ -566,7 +587,7 @@ BLKNO lowdup(originfo)
 	while (scan)
 	{
 		/* allocate a new blklist block, for the new buffer */
-		dupblklist = sesalloc(0);
+		dupblklist = sesalloc(0, SES_BLKLIST);
 
 		/* link the new blklist block into the new buffer's list */
 		if (prevdup != 0)
@@ -594,7 +615,7 @@ BLKNO lowdup(originfo)
 		for (i = 0; i < SES_MAXBLKLIST && blk->blklist.blk[i].blkno; i++)
 		{
 			/* Increment allocation count on the chars block */
-			(void)sesalloc(blk->blklist.blk[i].blkno);
+			(void)sesalloc(blk->blklist.blk[i].blkno, SES_CHARS);
 		}
 		assert(i == SES_MAXBLKLIST || !blk->blklist.next);
 
@@ -624,6 +645,9 @@ void lowfree(bufinfo)
 	BLK	*blk;
 	int	i;
 
+#if LINECACHE
+	clobbercache(bufinfo);
+#endif
 	/* Lock the superblock for writing */
 	scan = seslock(0, True, SES_SUPER);
 	assert(scan == 0);
@@ -654,8 +678,15 @@ void lowfree(bufinfo)
 		blk->super2.buf[i] = 0;
 	}
 	
-	/* Unlock the superblock for writing */
+	/* Unlock the superblock for writing, and then flush it to disk.
+	 * Flushing it is important because we're about to free the blocks
+	 * from this buffer, and those blocks are likely to be reallocated
+	 * right away for some other purpose.  If elvis crashed at that time
+	 * and the super block had never been flushed, then the session file
+	 * might be too scrambled for elvis to read it successfully.
+	 */
 	sesunlock(scan, True);
+	sesflush(scan);
 
 	/* Lock the bufinfo block */
 	(void)seslock(bufinfo, False, SES_BUFINFO);
@@ -690,8 +721,54 @@ void lowfree(bufinfo)
 	sesfree(bufinfo);
 }
 
+/* Change the title of a buffer */
+void lowtitle(bufinfo, title)
+	_BLKNO_	bufinfo;	/* bufinfo block whose title is to be changed */
+	CHAR	*title;		/* the new title */
+{
+	BLK	*blk;
+#ifdef DEBUG_SESSION
+	BLKNO	blkno;
+
+	blkno = seslock(bufinfo, True, SES_BUFINFO);
+	assert(blkno == bufinfo);
+#else
+	(void)seslock(bufinfo, True, SES_BUFINFO);
+#endif
+
+	blk = sesblk(bufinfo);
+	assert(blk);
+	CHARncpy(blk->bufinfo.name, title, (size_t)SES_MAXBUFINFO);
+	sesunlock(bufinfo, True);
+}
+
 /******************************************************************************/
 
+#if LINECACHE
+static struct
+{
+	BLKNO	bufinfo;/* bufinfo of a known line, or 0 for none */
+	long	lineno;	/* line number */
+	long	offset;	/* offset of the start of the line */
+} linecache[LINECACHE];
+static int lineidx;	/* index into linecache[] of slot to recycle */
+#endif /* LINECACHE */
+
+
+#if LINECACHE
+static void clobbercache(dst)
+	_BLKNO_	dst;	/* bufinfo of a changed buffer */
+{
+	register int	i;
+
+	for (i = 0; i < LINECACHE; i++)
+#if 0
+		if (linecache[i].bufinfo == dst)
+#endif
+			linecache[i].bufinfo = 0;
+}
+#endif /* LINECACHE */
+
 /* convert a line number to an offset, for a given buffer */
 long lowline(bufinfo, lineno)
 	_BLKNO_	bufinfo;	/* BUFINFO of the buffer */
@@ -702,6 +779,19 @@ long lowline(bufinfo, lineno)
 	BLKNO	next;	/* the next blklist block, after this one */
 	long	offset;	/* total offset seen */
 	int	i;	/* used for scanning through blklist.blk[] */
+#if LINECACHE
+	long	origline;/* a copy of lineno */
+
+	/* check the cache first */
+	for (i = 0; i < LINECACHE; i++)
+	{
+		if (linecache[i].bufinfo == bufinfo && linecache[i].lineno == lineno)
+		{
+			return linecache[i].offset;
+		}
+	}
+	origline = lineno;
+#endif /* LINECACHE */
 
 	/* outside this function, line numbers start with 1, but in here they start at 0 */
 	lineno--;
@@ -746,6 +836,14 @@ long lowline(bufinfo, lineno)
 					}
 				}
 
+#if LINECACHE
+				/* stuff information into the cache */
+				linecache[lineidx].bufinfo = bufinfo;
+				linecache[lineidx].offset = offset;
+				linecache[lineidx].lineno = origline;
+				lineidx = (lineidx + 1) % LINECACHE;
+#endif /* LINECACHE */
+
 				/* return the offset */
 				sesunlock(next, False);
 				return offset;
@@ -797,12 +895,12 @@ BLKNO lowoffset(bufinfo, offset, left, right, lptr, linenum)
 	LBLKNO	*lptr;		/* output: logical block number of CHARS block */
 	long	*linenum;	/* output: line number */
 {
-	BLK	*blk;	/* contents of a bufinfo block or blklist block */
 	BLKNO	blklist;/* used for stepping from one blklist to next */
 	BLKNO	next;	/* the next blklist block, after this one */
 	LBLKNO	lblkno;	/* counts overall LBLKNO of the given offset */
 	long	lnum;	/* counts newlines */
-	int	i;	/* used for scanning through blklist.blk[] */
+	register int	i;	/* used for scanning through blklist.blk[] */
+	register BLK	*blk;	/* contents of a bufinfo or blklist block */
 
 	/* treat negative offsets as 0 */
 	if (offset < 0)
@@ -892,10 +990,10 @@ BLKNO lowoffset(bufinfo, offset, left, right, lptr, linenum)
  * the number of lines in the buffer (i.e., the number of newline characters
  * in the "new" string.
  */
-long lowinsert(dst, dsttop, new, newlen)
+long lowinsert(dst, dsttop, newp, newlen)
 	_BLKNO_	dst;	/* BUFINFO of lowbuf */
 	long	dsttop;	/* offset where insertion should begin */
-register CHAR	*new;	/* new text to be inserted */
+register CHAR	*newp;	/* new text to be inserted */
 	long	newlen;	/* length of new text */
 {
 	BLKNO	blkno;
@@ -904,7 +1002,7 @@ register CHAR	*new;	/* new text to be inserted */
 	LBLKNO	lblkno;
 	BLK	*biblk;
 	BLK	*cblk;
-	COUNT	nlines;
+	int	nlines;
 register BLKNO	newblkno;	/* a new block */
 	BLKNO	blklist;
 	BLK	*newblk;
@@ -914,6 +1012,10 @@ register int	i;
 
 	safeinspect();
 
+#if LINECACHE
+	clobbercache(dst);
+#endif
+
 	/* figure out where insertion should begin */
 	blkno = lowoffset(dst, dsttop, &left, &right, &lblkno, (long *)0);
 
@@ -934,8 +1036,8 @@ register int	i;
 		nlines = 0;
 		for (i = 0; i < newlen; i++)
 		{
-			cblk->chars.chars[i + left] = new[i];
-			if (new[i] == '\n') nlines++;
+			cblk->chars.chars[i + left] = newp[i];
+			if (newp[i] == '\n') nlines++;
 		}
 
 		/* unlock that block, and update its nchars and nlines */
@@ -948,7 +1050,7 @@ register int	i;
 		if (blkno != 0 && left != 0 && right != 0)
 		{
 			/* allocate a new block */
-			newblkno = seslock(sesalloc(0), True, SES_CHARS);
+			newblkno = seslock(sesalloc(0, SES_CHARS), True, SES_CHARS);
 			newblk = sesblk(newblkno);
 
 			/* copy the second half of the characters into it */
@@ -981,8 +1083,8 @@ register int	i;
 			     i < o_blkfill && newlen > 0;
 			     i++, j++, newlen--)
 			{
-				cblk->chars.chars[i] = *new;
-				if (*new++ == '\n')
+				cblk->chars.chars[i] = *newp;
+				if (*newp++ == '\n')
 				{
 					nlines++;
 					totlines++;
@@ -1006,15 +1108,15 @@ register int	i;
 		while (newlen > 0)
 		{
 			/* allocate a new block */
-			newblkno = seslock(sesalloc(0), True, SES_CHARS);
+			newblkno = seslock(sesalloc(0, SES_CHARS), True, SES_CHARS);
 			newblk = sesblk(newblkno);
 
 			/* copy text into it */
 			nlines = 0;
 			for (i = 0; i < o_blkfill && newlen > 0; i++, newlen--)
 			{
-				newblk->chars.chars[i] = *new;
-				if (*new++ == '\n')
+				newblk->chars.chars[i] = *newp;
+				if (*newp++ == '\n')
 				{
 					nlines++;
 				}
@@ -1066,6 +1168,10 @@ long lowdelete(dst, dsttop, dstbottom)
 
 	safeinspect();
 
+#if LINECACHE
+	clobbercache(dst);
+#endif
+
 	/* lock the bufinfo block, and fetch it */
 	(void)seslock(dst, True, SES_BUFINFO);
 	biblk = sesblk(dst);
@@ -1073,6 +1179,7 @@ long lowdelete(dst, dsttop, dstbottom)
 	/* If both ends are in the same block */
 	first = lowoffset(dst, dsttop, &firstleft, &firstright, &firstlblk, (long *)0);
 	last = lowoffset(dst, dstbottom, &lastleft, &lastright, &lastlblk, (long *)0);
+
 	if (first == last || (firstlblk + 1 == lastlblk && lastleft == 0))
 	{
 		safeinspect();
@@ -1211,17 +1318,17 @@ long lowdelete(dst, dsttop, dstbottom)
 /* Replace one chunk of text with another, and return the change in the
  * number of lines.
  */
-long lowreplace(dst, dsttop, dstbottom, new, newlen)
+long lowreplace(dst, dsttop, dstbottom, newp, newlen)
 	_BLKNO_	dst;		/* BUFINFO of the lowbuf */
 	long	dsttop;		/* offset where replacement begins */
 	long	dstbottom;	/* offset of end of old text */
-	CHAR	*new;		/* new text */
+	CHAR	*newp;		/* new text */
 	long	newlen;		/* length of new text */
 {
 	long	totlines = 0;	/* total change in line count */
 
 	totlines += lowdelete(dst, dsttop, dstbottom);
-	totlines += lowinsert(dst, dsttop, new, newlen);
+	totlines += lowinsert(dst, dsttop, newp, newlen);
 	return totlines;
 }
 
@@ -1244,29 +1351,34 @@ long lowpaste(dst, dsttop, src, srctop, srcbottom)
 	safeinspect();
 
 	/* destination can't be in the middle of the source */
-	assert(src != dst || dsttop <= srctop || dsttop > srcbottom);
+	assert(src != dst || dsttop <= srctop || dsttop >= srcbottom);
 
-	/* Small problem: If the destination is in the same buffer as the source,
-	 * and happens to be located before the source in the same block, then
-	 * when inserting the new text we'll be altering the block cache which
-	 * could messes up our pointers.
+#if LINECACHE
+	clobbercache(dst);
+#endif
+
+	/* Small problem: If the destination is in the same buffer as the
+	 * source, and happens to be located before the source in the same
+	 * block, then when inserting the new text we'll be altering the block
+	 * cache which messes up our pointers.
 	 *
 	 * Solution: If source & dest are in the same buffer, and dest comes
 	 * before source, then we need to use a private copy of each source
 	 * block.
 	 */
-	if (src == dst && src <= dst)
+	if (dst == src && dsttop <= srctop)
 	{
-		tmpblk = safealloc((int)o_blksize, 1);
+		tmpblk = (BLK *)safealloc((int)o_blksize, 1);
 	}
 
 	/* we'll do this in chunks -- transfering text from each block of the
-	 * source into the destination, until all text has been teansferred.
+	 * source into the destination, until all text has been transferred.
 	 */
 	while (srctop < srcbottom)
 	{
 		/* read the text block, and find the start of source text in it */
 		blkno = lowoffset(src, srctop, &left, &right, (LBLKNO *)0, (long *)0);
+
 		seslock(blkno, False, SES_CHARS);
 		blk = sesblk(blkno);
 		if (tmpblk)
@@ -1286,7 +1398,7 @@ long lowpaste(dst, dsttop, src, srctop, srcbottom)
 		}
 
 		/* Copy the chunk of text into the destination. */
-		totlines += lowinsert(dst, dsttop, &blk->chars.chars[left], right);
+		totlines += lowinsert(dst, dsttop, &blk->chars.chars[left], (long)right);
 		srctop += right;
 		dsttop += right;
 
