@@ -1,21 +1,29 @@
 /* lp.c */
 /* Copyright 1995 by Steve Kirkendall */
 
-char id_lp[] = "$Id: lp.c,v 2.18 1996/05/24 00:56:28 steve Exp $";
+char id_lp[] = "$Id: lp.c,v 2.28 1997/11/02 18:54:45 steve Exp $";
 
 /* This file contains generic printing code. */
 
 #include "elvis.h"
+#ifdef FEATURE_LPR
 
 #if USE_PROTOTYPES
 static LPTYPE *findtype(char *name);
+#if defined (GUI_WIN32)
+static void dummyprt(_CHAR_ ch);
+#endif
 static void prtchar(_CHAR_ ch);
 static void draw(CHAR *p, long qty, _char_ font, long offset);
 #endif
 
 
 /* This is a list of all known printer types.  The first one is the default. */
-static LPTYPE	*alltypes[] = {&lpepson, &lppana, &lpibm, &lphp, &lpdumb, &lpcr, &lpbs, &lpps, &lpps2};
+static LPTYPE	*alltypes[] = {&lpepson, &lppana, &lpibm, &lphp, &lpdumb, &lpcr, &lpbs, &lpps, &lpps2,
+#if defined (GUI_WIN32)
+ &lpwindows,
+#endif
+};
 static LPTYPE	*type;		/* type of printer being used at the moment */
 
 /* Page formatting variables */
@@ -24,6 +32,7 @@ static int	linesleft;	/* number of usable lines left on this page */
 static int	column;		/* column number, used to implement wrapping */
 static WINDOW	prwin;		/* window doing the printing */
 static BOOLEAN	anytext;	/* False if blank page, True if any text */
+static long	lnum;		/* line number of current line */
 
 /* Output buffering variables */
 static CHAR	*iobuf;		/* pointer to the I/O buffer */
@@ -57,6 +66,12 @@ static LPTYPE *findtype(name)
 	return NULL;
 }
 
+#if defined (GUI_WIN32)
+static void dummyprt(ch)
+	_CHAR_	ch;	/* character to be sent to printer */
+{
+}
+#endif
 /* This function copies a single character to the printer (or file).
  * Internally, is uses a large I/O buffer for the sake of efficiency.
  */
@@ -85,6 +100,7 @@ static void draw(p, qty, font, offset)
 	WINDOW	win;
 	long	delta;
 	CHAR	ch;
+	char	lnumstr[20];
 
 	/* A negative qty indicates that the character *p is to be repeated.
 	 */
@@ -116,6 +132,17 @@ static void draw(p, qty, font, offset)
 			}
 		}
 
+		/* maybe output a line number */
+		if (column == 0		/* at start of line */
+		 && o_lpnumber		/* "lpnumber" option is set */
+		 && offset >= -1	/* not doing header or line number */
+		 && ch != '\f')		/* character isn't formfeed */
+		{
+			/* output a line number */
+			sprintf(lnumstr, "%6ld  ", lnum);
+			draw(toCHAR(lnumstr), 8, 'n', -2L);
+		}
+
 		/* if line is too long, then wrap it or clip it */
 		if (o_lpcolumns > 0 && column >= o_lpcolumns && ch != '\f' && ch != '\n')
 		{
@@ -208,6 +235,9 @@ RESULT lp(win, top, bottom, force)
 {
 	long	oldcurs;
 	MARK	next;
+	CHAR	*origdisplay = NULL;
+	char	*out;
+	char	rwa;
 	
 	/* convert the value of o_lptype to an LPTYPE pointer */
 	type = findtype(tochar8(o_lptype));
@@ -217,16 +247,38 @@ RESULT lp(win, top, bottom, force)
 		return RESULT_ERROR;
 	}
 
+	/* Switch to the bufdisplay display mode, if not that way already */
+	if (CHARcmp(o_display(win), o_bufdisplay(markbuffer(top))))
+	{
+		origdisplay = CHARdup(o_display(win));
+		if (!dispset(win, tochar8(o_bufdisplay(markbuffer(top)))))
+		{
+			safefree(origdisplay);
+			return RESULT_ERROR;
+		}
+	}
+
 	/* Call the mode's setup function.  Pretend the cursor is at the
 	 * top of the print region, so setup() doesn't try to scroll.
 	 */
-	next = (*win->md->setup)(top, markoffset(top), bottom, win->mi);
+	next = (*win->md->setup)(win, top, markoffset(top), bottom, win->mi);
 
 	/* open file or filter program */
-	assert(o_lpout);
-	if (!ioopen(tochar8(o_lpout), 'w', True, force, False))
+	if (type->spooled)
 	{
-		return RESULT_ERROR;
+		rwa = 'w';
+		out = tochar8(o_lpout);
+		if (!out || !*out)
+		{
+			msg(MSG_ERROR, "must set lpout");
+			goto Error;
+		}
+		if (out[0] == '>' && out[1] == '>')
+			out += 2, rwa = 'a';
+		if (!ioopen(out, rwa, True, force, 'b'))
+		{
+			goto Error;
+		}
 	}
 
 	/* allocate the I/O buffer */
@@ -235,7 +287,12 @@ RESULT lp(win, top, bottom, force)
 	ionext = 0;
 
 	/* initialize the LP driver */
-	(*type->before)(type->minorno, prtchar);
+#if defined (GUI_WIN32)
+	if (strcmp (o_lptype, "windows") == 0)
+	    (*type->before)(type->minorno, dummyprt);
+	else
+#endif
+		(*type->before)(type->minorno, prtchar);
 
 	/* start the first page */
 	pagenum = 1;
@@ -256,6 +313,7 @@ RESULT lp(win, top, bottom, force)
 	marksetoffset(win->cursor, o_bufchars(markbuffer(win->cursor)));
 	while (markoffset(next) < markoffset(bottom))
 	{
+		lnum = markline(next);
 		next = (*win->md->image)(win, next, win->mi, draw);
 	}
 	marksetoffset(win->cursor, oldcurs);
@@ -270,12 +328,24 @@ RESULT lp(win, top, bottom, force)
 	}
 	safefree(iobuf);
 	msg(MSG_INFO, "[d]$1 pages", (long)pagenum);
-	if (ioclose())
+	if (type->spooled && ioclose())
+		goto Error;
+
+	/* clean up and return success */
+	if (origdisplay)
 	{
-		return RESULT_COMPLETE;
+		(void)dispset(win, tochar8(origdisplay));
+		safefree(origdisplay);
 	}
-	else
+	return RESULT_COMPLETE;
+
+Error:
+	/* clean up and return failure */
+	if (origdisplay)
 	{
-		return RESULT_ERROR;
+		(void)dispset(win, tochar8(origdisplay));
+		safefree(origdisplay);
 	}
+	return RESULT_ERROR;
 }
+#endif /* FEATURE_LPR */
