@@ -1,7 +1,7 @@
 /* move.c */
 /* Copyright 1995 by Steve Kirkendall */
 
-char id_move[] = "$Id: move.c,v 2.35 1996/05/30 17:39:04 steve Exp $";
+char id_move[] = "$Id: move.c,v 2.43 1997/10/05 17:26:01 steve Exp $";
 
 #include "elvis.h"
 
@@ -162,17 +162,30 @@ RESULT m_front(win, vinf)
  * The number is given in the "count" field; if no number is given, then either
  * move to the last line, show buffer statistics, or move to matching bracket,
  * respectively.
+ * 
+ * Support for #if/#else/#endif derived from Antony Bowesman's contributed code.
+ *  When a # is found the next two characters are matched.  If an 'if' is 
+ *  found then the search is set forwards.  If an 'en' is found the search
+ *  goes backwards.  if an 'el' (For else or elif) the search direction
+ *  proceeds in the opposite direction of the last direction used in else
+ *  match.
  */
 RESULT m_absolute(win, vinf)
 	WINDOW	win;	/* window where command was typed */
 	VIINFO	*vinf;	/* information about the command */
 {
 	BUFFER	buf = markbuffer(win->state->cursor);
-	CHAR	match;	/* the parenthesis we want (for <%> command) */
-	CHAR	nest = 0;/* the starting parenthesis */
-	CHAR	*cp;	/* used for scanning through text */
-	long	count;	/* nesting depth */
-	EXINFO	xinfb;	/* an ex command */
+	CHAR	prepchar;	/* preprocessor character, usually '#' */
+	CHAR	match;		/* the parenthesis we want (for <%> command) */
+	CHAR	nest = 0;	/* the starting parenthesis */
+	CHAR	*cp, *cp2;	/* used for scanning through text */
+	long	count;		/* nesting depth */
+	EXINFO	xinfb;		/* an ex command */
+	BOOLEAN	fwd;		/* forward search? (else backward) */
+#ifdef DISPLAY_SYNTAX
+ static	BOOLEAN	fwdelse;	/* last "else" move direction */
+	char	prepname[3];	/* first two chars of preprocessor directive */
+#endif
 
 	assert(vinf->command == 'G' || vinf->command == ELVCTRL('G') || vinf->command == '%');
 
@@ -216,11 +229,16 @@ RESULT m_absolute(win, vinf)
 	  case '%':
 		if (!vinf->count)
 		{
+#ifdef DISPLAY_SYNTAX
+			/* find the preprocessor character, if any */
+			prepchar = dmspreprocessor(win);
+#endif
+
 			/* search forward within line for one of "[](){}" */
 			for (match = '\0', scanalloc(&cp, win->state->cursor); !match;)
 			{
 				/* if hit end-of-line or end-of-buffer without
-				 * finding a parenthesis, then fail.
+				 * finding a matchable character, then fail.
 				 */
 				if (!cp || *cp == '\n')
 				{
@@ -228,23 +246,99 @@ RESULT m_absolute(win, vinf)
 					return RESULT_ERROR;
 				}
 
-				/* if parenthesis, great! else keep looking */
-				switch (*cp)
+				/* is this a matchable char? */
+				nest = *cp;
+#ifdef DISPLAY_SYNTAX
+				if (prepchar && nest == prepchar)
 				{
-				  case '[':	match = ']';	break;
-				  case ']':	match = '[';	break;
-				  case '(':	match = ')';	break;
-				  case ')':	match = '(';	break;
-				  case '{':	match = '}';	break;
-				  case '}':	match = '{';	break;
-				  default:	scannext(&cp);
+					/* PREPROCESSOR DIRECTIVE */
+
+					/* skip whitespace, get directive name*/
+					scandup(&cp2, &cp);
+					do
+					{
+						scannext(&cp2);
+					} while (cp2 && (*cp2 == ' ' || *cp2 == '\t'));
+					memset(prepname, 0, sizeof prepname);
+					if (cp2)
+					{
+						prepname[0] = tolower(*cp2);
+						scannext(&cp2);
+						if (cp2)
+							prepname[1] = tolower(*cp2);
+					}
+					scanfree(&cp2);
+
+					/* choose direction from name */
+					if (!strcmp(prepname, "if"))
+					{
+						fwd = fwdelse = True;
+						match = prepchar;
+					}
+					else if (!strcmp(prepname, "en"))
+					{
+						fwd = fwdelse = False;
+						match = prepchar;
+					}
+					else if (!strcmp(prepname, "el"))
+					{
+						fwd = fwdelse;
+						match = prepchar;
+					}
+					else
+					{
+						/* UNRECOGNIZED DIRECTIVE */
+						scannext(&cp);
+					}
+				}
+				else
+#endif /* defined(DISPLAY_SYNTAX) */
+				{
+					/* look in matchchar option... */
+					for (cp2 = o_matchchar, fwd = True;
+					     cp2 && *cp2 && *cp2 != nest;
+					     cp2++, fwd = (BOOLEAN)!fwd)
+					{
+					}
+					if (!cp2 || !*cp2)
+					{
+						/* NOT MATCHABLE */
+						scannext(&cp);
+						continue;
+					}
+
+					/* figure out what matching char is */
+					if (!fwd)
+						/* MATCH PREVIOUS IN LIST */
+						match = cp2[-1];
+					else if (cp2[1])
+						/* MATCH FOLLOWING IN LIST */
+						match = cp2[1];
+					else
+						/* MATCH SELF */
+						match = nest;
+
+					/* if nest=match, choose a direction */
+					if (nest == match)
+					{
+						scandup(&cp2, &cp);
+						for (fwd = True;
+						     cp2 && *cp2 != '\n';
+						     scannext(&cp2))
+						{
+							if (*cp2 == nest)
+								fwd = (BOOLEAN)!fwd;
+						}
+						scanfree(&cp2);
+					}
 				}
 			}
 			assert(cp != NULL);
-			nest = *cp;
+			if (prepchar != match)
+				prepchar = '\0';
 
 			/* search forward or backward for match */
-			if (match == '(' || match == '[' || match == '{')/*)]}*/
+			if (!fwd)
 			{
 				/* search backward */
 				for (count = 1; count > 0; )
@@ -256,10 +350,32 @@ RESULT m_absolute(win, vinf)
 					}
 
 					/* check the char */
+#ifdef DISPLAY_SYNTAX
+					if (prepchar && *cp == prepchar)
+					{
+						if ((prepname[0] == 'i' && prepname[1] == 'f')
+						  || (count == 1 && prepname[0] == 'e' && prepname[1] == 'l'))
+							count--;
+						else if (prepname[0] == 'e' && prepname[1] == 'n')
+							count++;
+					}
+					else
+#endif
 					if (*cp == match)
 						count--;
 					else if (*cp == nest)
 						count++;
+
+#ifdef DISPLAY_SYNTAX
+					/* for benefit of preprocessor, shift
+					 * non-whitespace chars into prepname[]
+					 */
+					if (!isspace(*cp))
+					{
+						prepname[1] = prepname[0];
+						prepname[0] = tolower(*cp);
+					}
+#endif
 				}
 			}
 			else
@@ -274,7 +390,35 @@ RESULT m_absolute(win, vinf)
 					}
 
 					/* check the char */
-					if (*cp == match)
+#ifdef DISPLAY_SYNTAX
+					if (prepchar && *cp == prepchar)
+					{
+						/* peek ahead for prepname */
+						scandup(&cp2, &cp);
+						do
+						{
+							scannext(&cp2);
+						} while (cp2 && (*cp2 == ' ' || *cp2 == '\t'));
+						memset(prepname, 0, sizeof prepname);
+						if (cp2)
+						{
+							prepname[0] = tolower(*cp2);
+							scannext(&cp2);
+							if (cp2)
+								prepname[1] = tolower(*cp2);
+						}
+						scanfree(&cp2);
+
+						/* check for nesting, etc */
+						if (prepname[0] == 'e'
+						 && (prepname[1] == 'n'
+						    || (count == 1 && prepname[1] == 'l')))
+							count--;
+						else if (prepname[0] == 'i' && prepname[1] == 'f')
+							count++;
+					}
+#endif /* defined(DISPLAY_SYNTAX) */
+					else if (*cp == match)
 						count--;
 					else if (*cp == nest)
 						count++;
@@ -303,7 +447,7 @@ RESULT m_absolute(win, vinf)
 			 * I'm slightly careful here to avoid overflowing
 			 * the long int which stores the offset.
 			 */
-			if (o_bufchars(buf) > 1000000)
+			if (o_bufchars(buf) > 1000000L)
 			{
 				marksetoffset(win->state->cursor,
 					(o_bufchars(buf) / 100) * vinf->count);
@@ -705,11 +849,11 @@ RESULT	m_scroll(win, vinf)
 	  case ELVCTRL('Y'):
 		win->di->topline = markoffset((*win->md->move)(win, marktmp(tmp, markbuffer(win->cursor), win->di->topline), -vinf->count, 0, True));
 		win->di->bottomline = markoffset((*win->md->move)(win, marktmp(tmp, markbuffer(win->cursor), win->di->bottomline), -vinf->count, 0, True));
-		marksetoffset(win->cursor, markoffset(dispmove(win, 0, 0)));
+		marksetoffset(win->cursor, markoffset(dispmove(win, 0, win->wantcol)));
 		if (markoffset(win->cursor) >= win->di->bottomline)
 		{
 			marksetoffset(win->cursor, win->di->bottomline);
-			marksetoffset(win->cursor, markoffset(dispmove(win, -1, 0)));
+			marksetoffset(win->cursor, markoffset(dispmove(win, -1, win->wantcol)));
 		}
 		break;
 
@@ -719,6 +863,7 @@ RESULT	m_scroll(win, vinf)
 		if (markoffset(win->cursor) < win->di->topline)
 		{
 			marksetoffset(win->cursor, win->di->topline);
+			marksetoffset(win->cursor, markoffset(dispmove(win, 0, win->wantcol)));
 		}
 		break;
 
@@ -799,11 +944,27 @@ RESULT	m_csearch(win, vinf)
 	static CHAR	prevcmd;	/* previous command, for <,> and <;> */
 	static CHAR	prevtarget;	/* previous target character */
 	CHAR		*cp;		/* used for scanning text */
+	long		front, end;	/* endpoints of the line */
+	long		offset;		/* offset of current char in line */
 
 	DEFAULT(1);
 
 	assert(strchr("fFtT,;", vinf->command));
 
+	/* Find the endpoints of this line.  Note that we need to be careful
+	 * about the end of the line, because we don't want to allow the
+	 * newline character to be deleted.
+	 */
+	front = markoffset(dispmove(win, 0, 0));
+	if (win->state->acton)
+	{
+		end = markoffset((*dmnormal.move)(win, win->state->cursor, 0, INFINITY, True));
+	}
+	else
+	{
+		end = markoffset((*win->md->move)(win, win->state->cursor, 0, INFINITY, True));
+	}
+
 	/* comma and semicolon recall the previous character search */
 	if (vinf->command == ';' || vinf->command == ',')
 	{
@@ -839,12 +1000,12 @@ RESULT	m_csearch(win, vinf)
 	}
 
 	/* Which way should we scan?  Forward or backward? */
+	offset = markoffset(win->state->cursor);
+	(void)scanalloc(&cp, win->state->cursor);
 	if (islower(vinf->command))
 	{
 		/* scan forward */
-		for (scanalloc(&cp, win->state->cursor);
-		     vinf->count > 0 && scannext(&cp) && *cp != '\n';
-		     )
+		while (vinf->count > 0 && scannext(&cp) && ++offset <= end)
 		{
 			if (*cp == vinf->key2)
 			{
@@ -855,9 +1016,7 @@ RESULT	m_csearch(win, vinf)
 	else
 	{
 		/* scan backward */
-		for (scanalloc(&cp, win->state->cursor);
-		     vinf->count > 0 && scanprev(&cp) && *cp != '\n';
-		     )
+		while (vinf->count > 0 && scanprev(&cp) && --offset >= front)
 		{
 			if (*cp == vinf->key2)
 			{
@@ -867,7 +1026,7 @@ RESULT	m_csearch(win, vinf)
 	}
 
 	/* if hit EOF or newline, then fail */
-	if (!cp || *cp == '\n')
+	if (!cp || offset < front || offset > end)
 	{
 		scanfree(&cp);
 		return RESULT_ERROR;
@@ -1291,11 +1450,16 @@ RESULT m_fsentence(win, vinf)
 	VIINFO	*vinf;	/* information about the command */
 {
 	BOOLEAN	ending;	/* have we seen at least one sentence ender? */
+	BOOLEAN	didpara;/* between paragraph and first sentence in paragraph */
 	int	spaces;	/* number of spaces seen so far */
 	CHAR	*cp;	/* used for scanning through text */
 	CHAR	*end;	/* characters that end a sentence */
 	CHAR	*quote;	/* quote/parenthesis character that may appear at end */
+	CHAR	oper;	/* operator command, or '\0' */
 	long	newline;/* offset of first newline in trailing whitespace */
+	long	para;
+	long	offset;
+	long	count;
 
 	DEFAULT(1);
 
@@ -1303,22 +1467,108 @@ RESULT m_fsentence(win, vinf)
 	end = o_sentenceend ? o_sentenceend : toCHAR(".?!");
 	quote = o_sentencequote ? o_sentencequote : toCHAR("\")]");
 
+	count = vinf->count;
+	oper = vinf->oper;
+
+	/* detect whether we're at the start of a paragraph */
+	offset = markoffset(win->state->cursor);
+	scanalloc(&cp, win->state->cursor);
+	if (*cp != '\n')
+		scanprev(&cp);
+	else
+		while (cp && *cp == '\n')
+		{
+			scanprev(&cp);
+		}
+	if (cp)
+	{
+		marksetoffset(win->state->cursor, markoffset(scanmark(&cp)));
+		vinf->count = 1;
+		vinf->command = '}';
+		m_fsection(win, vinf);
+		para = markoffset(win->state->cursor);
+	}
+	else
+	{
+		para = 0;
+	}
+	marksetoffset(win->state->cursor, offset);
+	if (para == offset)
+		count++;
+	else if (para < offset)
+	{
+		/* find the next paragraph */
+		marksetoffset(win->state->cursor, markoffset(scanmark(&cp)));
+		vinf->count = 1;
+		vinf->command = '}';
+		m_fsection(win, vinf);
+		para = markoffset(win->state->cursor);
+	}
+
 	/* for each count... */
 	newline = -1;
-	scanalloc(&cp, win->state->cursor);
-	for (; cp && vinf->count > 0; vinf->count--)
+	scanseek(&cp, win->state->cursor);
+	for (; cp && count > 0; count--)
 	{
 		/* for each character in the sentence... */
-		for (ending = False, spaces = 0;
+		for (ending = didpara = False, spaces = 0;
 		     cp && (!ending || spaces < o_sentencegap || isspace(*cp));
-		     scannext(&cp))
+		     scannext(&cp), offset++)
 		{
+			/* if paragraph, then... */
+			if (offset == para)
+			{
+				/* if still more sentences to skip... */
+				if (count > 1)
+				{
+					/* count this as a sentence, and
+					 * arrange for next line to also be a
+					 * sentence.
+					 */
+					count--;
+					newline = -1;
+					ending = True;
+					if (*cp == '\n')
+					{
+						didpara = False;
+						spaces = o_sentencegap;
+					}
+					else
+					{
+						didpara = True;
+						spaces = 0;
+					}
+
+					/* oh, and we need to find the next
+					 * paragraph, too.
+					 */
+					marksetoffset(win->state->cursor, offset);
+					vinf->count = 1;
+					vinf->command = '}';/*{*/
+					if (m_fsection(win, vinf) == RESULT_COMPLETE)
+						para = markoffset(win->state->cursor);
+					else
+						para = -1;
+
+					continue;
+				}
+				else
+				{
+					/* we're here! */
+					break;
+				}
+			}
+
+			/* check the character */
 			if (*cp == '\n')
 			{
 				spaces = o_sentencegap;
+				didpara = False;
 				if (newline < 0)
 					newline = markoffset(scanmark(&cp));
 			}
+			else if (didpara)
+				/* skip characters in a ".P" line */;
 			else if (isspace(*cp))
 				spaces++;
 			else if (CHARchr(end, *cp))
@@ -1337,7 +1587,7 @@ RESULT m_fsentence(win, vinf)
 		 * trailing whitespace, then move the cursor to that newline;
 		 * else move the cursor to the start of the following sentence.
 		 */
-		if (vinf->oper && newline >= 0)
+		if (oper && newline >= 0)
 			marksetoffset(win->state->cursor, newline);
 		else
 			marksetoffset(win->state->cursor, markoffset(scanmark(&cp)));
@@ -1352,10 +1602,15 @@ RESULT m_bsentence(win, vinf)
 {
 	BOOLEAN	first;	/* True until we pass some mid-sentence stuff */
 	BOOLEAN	ending;	/* have we seen at least one sentence ender? */
+	BOOLEAN	anynext;/* any text seen on following line */
+	BOOLEAN	anythis;/* any text seen on this line */
 	int	spaces;	/* number of spaces seen so far */
 	CHAR	*cp;	/* used for scanning through text */
 	CHAR	*end;	/* characters that end a sentence */
 	CHAR	*quote;	/* quote/parenthesis character that may appear at end */
+	long	count;	/* sentences to move over */
+	long	para;	/* top of current paragraph */
+	long	offset;
 	RESULT	result;
 
 	DEFAULT(1);
@@ -1364,6 +1619,10 @@ RESULT m_bsentence(win, vinf)
 	end = o_sentenceend ? o_sentenceend : toCHAR(".?!");
 	quote = o_sentencequote ? o_sentencequote : toCHAR("\")]");
 
+	/* misc initialization */
+	anynext = anythis = False;
+	offset = markoffset(win->state->cursor);
+
 	/* NOTE: The "first" variable is used to handle the situation where
 	 * we start at the beginning of one sentence.  From here, we want
 	 * to go back one extra sentence-end; otherwise <(> would just move
@@ -1371,46 +1630,126 @@ RESULT m_bsentence(win, vinf)
 	 */
 	first = True;
 
-	/* for each count... */
+	/* Start scanning at the cursor location */
 	scanalloc(&cp, win->state->cursor);
-	for (; cp && vinf->count > 0; vinf->count--)
+	count = vinf->count;
+
+	/* Find the start of this paragraph.  That counts as a "sentence" */
+	vinf->command = '{';
+	vinf->count = 1;
+	para = (m_bsection(win, vinf) == RESULT_COMPLETE) ? markoffset(win->state->cursor) : -1;
+
+	/* For each count... */
+	for (; cp && count > 0; count--)
 	{
 		/* for each character in the sentence... */
-		for (ending = True, spaces = 0, scanprev(&cp);
-		     cp && (!ending || spaces < o_sentencegap || !CHARchr(end, *cp));
-		     scanprev(&cp))
+		for (ending = True, anythis = anynext = False, spaces = 0,
+			scanprev(&cp), offset = markoffset(scanmark(&cp));
+		     cp && offset != para &&
+			(!ending || spaces<o_sentencegap || !CHARchr(end,*cp));
+		     scanprev(&cp), offset--)
 		{
 			if (*cp == '\n')
-				spaces = o_sentencegap, ending = True;
+			{
+				spaces = o_sentencegap;
+				ending = True;
+				anynext = anythis;
+				anythis = False;
+			}
 			else if (isspace(*cp))
-				spaces++, ending = True;
+			{
+				spaces++,
+				ending = True;
+			}
 			else if (!CHARchr(quote, *cp))
-				first = ending = False, spaces = 0;
+			{
+				first = ending = False;
+				anythis = True;
+				spaces = 0;
+			}
+			else
+			{
+				anythis = True;
+			}
+		}
+
+		/* If this sentence is actually a paragraph start, and we
+		 * still have more sentences to move over, then find the
+		 * next higher paragraph.
+		 */
+		if (offset == para && count > 1)
+		{
+			vinf->count = 1;
+			para = (m_bsection(win, vinf) == RESULT_COMPLETE) ? markoffset(win->state->cursor) : -1;
 		}
 
 		/* If this sentence ender was encountered before any
 		 * mid-sentence text (i.e., if we started at the front of
 		 * a sentence) then we should go back one extra sentence-end.
 		 */
-		if (first)
+		if (first && offset != para)
 		{
-			vinf->count++;
+			count++;
 		}
 	}
 
+	/* If we hit a paragraph top which occurs on a blank line, then we
+	 * need to do a little extra processing because we exited the loop
+	 * a little too soon.
+	 */
+	if (offset == para && cp && *cp == '\n')
+	{
+		anynext = anythis;
+	}
+
 	/* did we find it? */
 	if (cp)
 	{
 		/* move the cursor to the start of the next sentence */
-		do
+		if (offset > para)
 		{
-			scannext(&cp);
-			assert(cp);
-		} while (isspace(*cp) || CHARchr(quote, *cp));
-		marksetoffset(win->state->cursor, markoffset(scanmark(&cp)));
+			/* found a sentence end -- move the cursor to the to
+			 * start of the following sentence.
+			 */
+			do
+			{
+				scannext(&cp);
+				assert(cp);
+			} while (isspace(*cp) || CHARchr(quote, *cp));
+			marksetoffset(win->state->cursor, markoffset(scanmark(&cp)));
+		}
+		else if (anynext)
+		{
+			/* bumped into a paragraph start after scanning some
+			 * sentence text -- move the cursor to the first text
+			 * on the next non-blank line.
+			 */
+			while (cp && *cp != '\n')
+			{
+				scannext(&cp);
+			}
+			while (cp && isspace(*cp))
+			{
+				scannext(&cp);
+			}
+			if (&cp)
+				marksetoffset(win->state->cursor, markoffset(scanmark(&cp)));
+			else
+			{
+				marksetoffset(win->state->cursor, o_bufchars(markbuffer(win->state->cursor)) - 2);
+				if (markoffset(win->state->cursor) < 0)
+					marksetoffset(win->state->cursor, 0);
+			}
+		}
+		else
+		{
+			/* found a paragraph start -- leave the cursor there */
+			offset = para;
+			marksetoffset(win->state->cursor, para);
+		}
 		result = RESULT_COMPLETE;
 	}
-	else if (vinf->count == 0)
+	else if (count == 0)
 	{
 		/* move the cursor to the first character of the buffer */
 		marksetoffset(win->state->cursor, 0);
