@@ -1,7 +1,7 @@
 /* operator.c */
 /* Copyright 1995 by Steve Kirkendall */
 
-char id_operator[] = "$Id: operator.c,v 2.31 1996/05/30 17:39:04 steve Exp $";
+char id_operator[] = "$Id: operator.c,v 2.39 1997/12/22 20:09:37 steve Exp $";
 
 #include "elvis.h"
 
@@ -10,6 +10,8 @@ char id_operator[] = "$Id: operator.c,v 2.31 1996/05/30 17:39:04 steve Exp $";
 static RESULT	filterenter(WINDOW win);
 #endif
 
+static MARKBUF	fromdup, todup;
+
 /* This function sends a portion of some buffer through an external filter
  * program, and replaces the old text with the output of the filter program.
  */
@@ -23,6 +25,13 @@ RESULT opfilter(from, to, prog)
 	MARKBUF	mark;	/* insertion point for replacement text */
 	CHAR	*p;	/* used for scanning through text, or ptr to input buf */
 
+	/* if no filter, then fail */
+	if (!prog || !*prog)
+	{
+		msg(MSG_ERROR, "no filter");
+		return RESULT_ERROR;
+	}
+
 	/* start the filter */
 	if (gui->prgopen
 		? !(*gui->prgopen)(tochar8(prog), True, True)
@@ -59,7 +68,7 @@ RESULT opfilter(from, to, prog)
 	}
 
 	/* read in the new text */
-	for (mark = *to, p = safealloc((int)o_blksize, sizeof(CHAR));
+	for (mark = *to, p = (CHAR *)safealloc((int)o_blksize, sizeof(CHAR));
 	     (nchars = prgread(p, (int)o_blksize)) > 0;
 	     markaddoffset(&mark, nchars))
 	    
@@ -91,15 +100,12 @@ static RESULT	filterenter(win)
 	CHAR	*cmd;	/* used for collecting the characters of the command line */
 	CHAR	*cp;	/* used for scanning the command-line buffer */
 	CHAR	*sub;	/* used for scanning a substitution value (for ! % #) */
-	char	*err;	/* error message (if this is an error) */
+	char	err;	/* error message (if this is an error) */
 	long	i;
 
 	assert(markoffset(win->state->top) <= markoffset(win->state->cursor));
 	assert(markoffset(win->state->cursor) <= markoffset(win->state->bottom));
 
-	/* Initialize "err" just to silence a compiler warning */
-	err = NULL;
-
 	/* skip over the '!' at the beginning of the command line, and any
 	 * whitespace.
 	 */
@@ -118,21 +124,19 @@ static RESULT	filterenter(win)
 	     cp != NULL && i > 1;
 	     scannext(&cp), i--)
 	{
+		err = *cp;
 		switch (*cp)
 		{
 		  case '!':
 			sub = o_previouscommand;
-			err = "no previous command";
 			break;
 
 		  case '%':
 			sub = o_filename(markbuffer(win->cursor));
-			err = "no file name";
 			break;
 
 		  case '#':
-			sub = o_previousfile;
-			err = "no previous file";
+			sub = buffilenumber(&cp);
 			break;
 
 		  case '\\':
@@ -154,7 +158,7 @@ static RESULT	filterenter(win)
 			(void)buildCHAR(&cmd, *cp);
 		else if (!sub)
 		{
-			msg(MSG_ERROR, err);
+			msg(MSG_ERROR, "[c]no value to substitute for $1", err);
 			goto Fail;
 		}
 		else
@@ -175,7 +179,14 @@ static RESULT	filterenter(win)
 		safefree(o_previouscommand);
 	o_previouscommand = cmd;
 
-	return RESULT_COMPLETE;
+	/* blank out the parsing info */
+	assert(win->state->pop && win->state->acton == win->state->pop);
+	viinitcmd((VIINFO *)win->state->pop->info);
+	win->state->pop->flags &= ~ELVIS_MORE; /* !!! why is this necessary? */
+
+	/* do the filtering */
+	assert(markbuffer(&fromdup) == markbuffer(&todup));
+	return opfilter(&fromdup, &todup, o_previouscommand);
 
 Fail:	if (cmd)
 		safefree(cmd);
@@ -188,13 +199,12 @@ Fail2:
 
 
 /* This function applies an operator to a chunk of text */
-RESULT operator(win, vinf, from, to)
+RESULT oper(win, vinf, from, to)
 	WINDOW	win;	/* where the command was typed */
 	VIINFO	*vinf;	/* information about the command */
 	MARK	from;	/* start of the affected text */
 	MARK	to;	/* end of the affected text */
 {
- static MARKBUF	fromdup, todup;
 	RESULT	result;
 	EXINFO	xinfb;
 	BOOLEAN	dot;
@@ -239,7 +249,6 @@ RESULT operator(win, vinf, from, to)
 	if (vinf->oper == '!' && !dot && !(win->state->flags & ELVIS_MORE))
 	{
 		statestratum(win, toCHAR(FILTER_BUF), vinf->oper, filterenter);
-		o_internal(markbuffer(win->state->cursor)) = True;
 		o_inputtab(markbuffer(win->state->cursor)) = 'f'; /* filename */
 
 		/* remember the range */
@@ -287,7 +296,15 @@ RESULT operator(win, vinf, from, to)
 			/* replace the text with previous input */
 			curs = cutput('.', win, from, False, True, True);
 
-			/* leave the cursor at the end of the replacement text */
+			/* if line mode, then append a newline */
+			if (vinf->tweak & TWEAK_LINE)
+			{
+				markaddoffset(curs, 1L);
+				bufreplace(curs, curs, toCHAR("\n"), 1);
+				markaddoffset(curs, -1L);
+			}
+
+			/* leave cursor at the end of the replacement text */
 			marksetoffset(win->state->cursor, markoffset(curs));
 		}
 		else
@@ -303,6 +320,7 @@ RESULT operator(win, vinf, from, to)
 	  case '>':
 		/* build a :< or :> ex command */
 		memset((char *)&xinfb, 0, sizeof xinfb);
+		xinfb.cmdname = (vinf->oper == '<') ? "<" : ">";
 		xinfb.command = (vinf->oper == '<') ? EX_SHIFTL : EX_SHIFTR;
 		xinfb.defaddr = *from;
 		xinfb.fromaddr = from;
