@@ -18,26 +18,69 @@ static int fd = -1; /* file descriptor of the session file */
  */
 BOOLEAN blkopen(BOOLEAN force, BLK *buf)
 {
-	char	sesname[16];
+	char	sesname[256];
 	char	*sespath;
+	int	len;
 	int	i = 1;
 
+	/* search through SESSIONPATH for a writable directory */
+	sespath = getenv("SESSIONPATH");
+	if (!sespath)
+		sespath = getenv("TMP");
+	if (!sespath)
+		sespath = "C:\\TMP;~;";
+	do
+	{
+		for (len = 0; *sespath && *sespath != ';'; sespath++)
+		{
+			if (len == 0 && *sespath == '~')
+			{
+				strcpy(sesname, tochar8(o_home));
+				len = strlen(sesname);
+			}
+			else
+			{
+				sesname[len++] = *sespath;
+			}
+		}
+		if (len == 0)
+		{
+			strcpy(sesname, ".");
+			len = 1;
+		}
+		if (sesname[len - 1] != '\\')
+			sesname[len++] = '\\';
+		strcpy(sesname + len, "elvis.tmp");
+	} while ((i = open(sesname, O_RDWR|O_CREAT, 0666)) < 0 && *sespath++);
+	if (i < 0)
+	{
+		msg(MSG_FATAL, "set SESSIONPATH to a writable directory");
+	}
+	close(i);
+	remove(sesname);
+
 	/* If no session file was explicitly requested, use the default */
 	if (!o_session)
 	{
 		if (o_recovering)
 			do
 			{
-				sprintf(sesname, DEFAULT_SESSION, i++);
-				sespath = dirpath(o_home, sesname);
-			} while (_access(sespath, 6) == -1 && errno == ENOENT);
+				sprintf(sesname + len, DEFAULT_SESSION, i++);
+
+				/* If the user wants to cancel, then fail */
+				if (chosengui->poll && (*chosengui->poll)(False))
+					return False;
+			} while (_access(sesname, 6) == -1 && errno == ENOENT);
 		else
 			do
 			{
-				sprintf(sesname, DEFAULT_SESSION, i++);
-				sespath = dirpath(o_home, sesname);
-			} while (_access(sespath, 0) == 0);
-		optpreset(o_session, CHARdup(toCHAR(sespath)), OPT_LOCK|OPT_FREE);
+				sprintf(sesname + len, DEFAULT_SESSION, i++);
+
+				/* If the user wants to cancel, then fail */
+				if (chosengui->poll && (*chosengui->poll)(False))
+					return False;
+			} while (_access(sesname, 0) == 0);
+		optpreset(o_session, CHARdup(toCHAR(sesname)), OPT_LOCK|OPT_FREE);
 		o_tempsession = !o_recovering;
 	}
 
@@ -139,5 +182,8 @@ void blkread(BLK *buf, _BLKNO_ blkno)
 /* Force changes out to disk */
 void blksync P_((void))
 {
-	/* !!! how? */
+    int     dupfd;
+
+    if ((dupfd = _dup (fd)) != -1)
+        _close (dupfd);
 }
