@@ -29,8 +29,8 @@
 /*----------------------------------------------------------------------------*/
 /* session file format                                                        */
 
-#define SESSION_MAGIC			0x0200DEAD
-#define SESSION_MAGIC_BYTESWAPPED	0xADDE0002
+#define SESSION_MAGIC			0x0200DEADL
+#define SESSION_MAGIC_BYTESWAPPED	0xADDE0002L
 
 /* These data types are used to represent physical and logical block numbers.
  * BLKNO is a physical block number; it is used to compute an offset into the
@@ -101,17 +101,27 @@ typedef union
 /*----------------------------------------------------------------------------*/
 /* low-level session file access functions                                    */
 
+BEGIN_EXTERNC
 extern void	sesopen P_((BOOLEAN force));
 extern void	sesclose P_((void));
 extern BLK	*sesblk P_((_BLKNO_));
 extern void	sesunlock P_((_BLKNO_ blkno, BOOLEAN forwrite));
 extern void	sesflush P_((_BLKNO_ blkno));
 extern void	sessync P_((void));
-extern BLKNO	sesalloc P_((_BLKNO_ blkno));
-extern void	sesfree P_((_BLKNO_ blkno));
+END_EXTERNC
 #ifdef DEBUG_SESSION
 # define seslock(b,f,t)	_seslock(__FILE__, __LINE__, b, f, t)
+# define sesalloc(b,t)	_sesalloc(__FILE__, __LINE__, b, t)
+# define sesfree(b)	_sesfree(__FILE__, __LINE__, b)
+BEGIN_EXTERNC
 extern BLKNO	_seslock P_((char *file, int line, _BLKNO_ blkno, BOOLEAN forwrite, BLKTYPE blktype));
+extern BLKNO	_sesalloc P_((char *file, int line, _BLKNO_ blkno, BLKTYPE blktype));
+extern void	_sesfree P_((char *file, int line, _BLKNO_ blkno));
+END_EXTERNC
 #else
+BEGIN_EXTERNC
 extern BLKNO	seslock P_((_BLKNO_ blkno, BOOLEAN forwrite, BLKTYPE blktype));
+extern BLKNO	sesalloc P_((_BLKNO_ blkno, BLKTYPE blktype));
+extern void	sesfree P_((_BLKNO_ blkno));
+END_EXTERNC
 #endif
