@@ -1,7 +1,7 @@
 /* verify.c */
 /* Copyright 1995 by Steve Kirkendall */
 
-char id_verify[] = "$Id: verify.c,v 1.4 1996/05/30 17:39:04 steve Exp $";
+char id_verify[] = "$Id: verify.c,v 1.8 1998/09/20 18:09:28 steve Exp $";
 
 /* This file contains a replacement for elvis' main() function.  The resulting
  * program will test each component of elvis.
@@ -10,13 +10,56 @@ char id_verify[] = "$Id: verify.c,v 1.4 1996/05/30 17:39:04 steve Exp $";
 #include <stdio.h>
 #include "elvis.h"
 
+#ifdef FEATURE_COMPLETE
+# error You must #undef FEATURE_COMPLETE in config.h to compile this program
+#endif
+#ifdef FEATURE_SHOWTAG
+# error You must #undef FEATURE_SHOWTAG in config.h to compile this program
+#endif
+
+
 static BLK blkbuf;
 
+#ifdef USE_PROTOTYPES
+extern int main(int argc, char **argv);
+static void init(void);
+extern BOOLEAN buffer(void);
+extern BOOLEAN lowbuf(void);
+extern BOOLEAN session(void);
+extern BOOLEAN options(void);
+extern BOOLEAN ostext(void);
+extern BOOLEAN osblock(void);
+extern BOOLEAN osdir(void);
+extern BOOLEAN sesmacro(void);
+extern WINDOW winofbuf(WINDOW win, BUFFER buf);
+extern WINDOW winalloc(GUIWIN *gw, OPTVAL *gvals, BUFFER buf, long rows, long columns);
+extern void msg(MSGIMP imp, char *terse, ...);
+extern void drawchar(CHAR *cp, long qty, _char_ font, long offset);
+extern void draw1ch(CHAR c, _char_ font, long offset);
+extern void drawopencomplete(WINDOW win);
+extern char *perm2str(DIRPERM perm);
+extern char *bool2str(BOOLEAN bool);
+extern void urlbytes(long totbytes);
+extern void urlclose(void);
+extern char *urllocal(char *url);
+extern BOOLEAN urlopen(char *url, BOOLEAN force, _char_ rwa);
+extern DIRPERM urlperm(char *url);
+extern int urlread(CHAR *buf, int bytes);
+extern BOOLEAN urlremote(char *url);
+extern int urlwrite(CHAR *buf, int bytes);
+extern CHAR *msgtranslate(char *word);
+extern TAG *tetag(CHAR *select);
+extern BUFFER cutbuffer(_CHAR_ cbname, BOOLEAN	create);
+#endif /* USE_PROTOTYPES */
 
 /*****************************************************************************
  * Start of stub functions
  */
 
+BOOLEAN exthenflag;
+long eventcounter;
+GUI dummygui, *chosengui = &dummygui;
+
 char *bool2str(BOOLEAN bool)
 {
 	return (bool ? "TRUE" : "FALSE");
@@ -32,11 +75,11 @@ char *perm2str(DIRPERM perm)
 	  case DIR_NEW:		return "DIR_NEW";
 	  case DIR_UNREADABLE:	return "DIR_UNREADABLE";
 	  case DIR_READONLY:	return "DIR_READONLY";
-	  case DIR_READWRITE:	return "DIR_READWRITE";
+	  default:		return "DIR_READWRITE";
 	}
 }
 
-void drawchar(CHAR c, char font, long offset)
+void draw1ch(CHAR c, _char_ font, long offset)
 {
 	static char prevfont = 'n';
 
@@ -53,6 +96,26 @@ void drawchar(CHAR c, char font, long offset)
 	putchar(c);
 }
 
+void drawchar(CHAR *cp, long qty, _char_ font, long offset)
+{
+	long	delta = 1;
+	if (qty < 0)
+	{
+		qty = -qty;
+		delta = 0;
+	}
+	while(--qty >= 0)
+	{
+		draw1ch(*cp, font, offset);
+		cp += delta;
+		offset += delta;
+	}
+}
+
+void drawopencomplete(WINDOW win)
+{
+}
+
 void msg(MSGIMP imp, char *terse, ...)
 {
 	fputs(terse, stderr);
@@ -62,6 +125,9 @@ void msg(MSGIMP imp, char *terse, ...)
 	}
 }
 
+void drawextext(WINDOW win, CHAR *text, int len){}
+int wincount(BUFFER buf){ return 0; }
+WINDOW windefault;
 WINDOW windows;
 WINDOW winalloc(GUIWIN *gw, OPTVAL *gvals, BUFFER buf, long rows, long columns)
 {
@@ -110,6 +176,77 @@ WINDOW winofbuf(WINDOW win, BUFFER buf)
 }
 
 
+RESULT experform(WINDOW win, MARK top, MARK bottom)
+{
+	return RESULT_ERROR;
+}
+BOOLEAN exparseaddress(CHAR **refp, EXINFO *xinf)
+{
+	return False;
+}
+void exfree(EXINFO *xinf)
+{
+}
+RESULT _viperform(WINDOW win)
+{
+	return RESULT_ERROR;
+}
+
+void urlbytes(long totbytes)
+{
+}
+
+void urlclose(void)
+{
+}
+
+char *urllocal(char *url)
+{
+	return url;
+}
+
+BOOLEAN urlopen(char *url, BOOLEAN force, _char_ rwa)
+{
+	return False;
+}
+
+DIRPERM urlperm(char *url)
+{
+	return dirperm(url);
+}
+
+int urlread(CHAR *buf, int bytes)
+{
+	return 0;
+}
+
+BOOLEAN urlremote(char *url)
+{
+	return False;
+}
+
+int urlwrite(CHAR *buf, int bytes)
+{
+	return 0;
+}
+
+CHAR *msgtranslate(char *word)
+{
+	return toCHAR(word);
+}
+
+TAG *tetag(CHAR *select)
+{
+	return NULL;
+}
+
+BUFFER cutbuffer(_CHAR_ cbname, BOOLEAN	create)
+{
+	abort();
+}
+
+
+
 /*
  * End of stubs
  *****************************************************************************
@@ -117,27 +254,27 @@ WINDOW winofbuf(WINDOW win, BUFFER buf)
  */
 
 BOOLEAN sesmacro(void)
-{
+{ 
 	BOOLEAN passed = True;
 
 	if ((char *)&blkbuf.super.buf[SES_MAXSUPER] > &blkbuf.sizetester[o_blksize])
 	{
-		printf("\tSES_MAXSUPER = %d\n", SES_MAXSUPER);
+		printf("\tSES_MAXSUPER = %ld\n", SES_MAXSUPER);
 		passed = False;
 	}
 	if ((char *)&blkbuf.bufinfo.name[SES_MAXBUFINFO] > &blkbuf.sizetester[o_blksize])
 	{
-		printf("\tSES_MAXBUFINFO = %d\n", SES_MAXBUFINFO);
+		printf("\tSES_MAXBUFINFO = %ld\n", SES_MAXBUFINFO);
 		passed = False;
 	}
 	if ((char *)&blkbuf.blklist.blk[SES_MAXBLKLIST] > &blkbuf.sizetester[o_blksize])
 	{
-		printf("\tSES_MAXBLKLIST = %d\n", SES_MAXBLKLIST);
+		printf("\tSES_MAXBLKLIST = %ld\n", SES_MAXBLKLIST);
 		passed = False;
 	}
 	if ((char *)&blkbuf.chars.chars[SES_MAXCHARS] > &blkbuf.sizetester[o_blksize])
 	{
-		printf("\tSES_MAXCHARS = %d\n", SES_MAXCHARS);
+		printf("\tSES_MAXCHARS = %ld\n", SES_MAXCHARS);
 		passed = False;
 	}
 
@@ -160,7 +297,7 @@ BOOLEAN osdir(void)
 
 	/* wildcard expansion */
 	printf("\tdiriswild(\"*.c\") = %s\n", bool2str(diriswild("*.c")));
-	tmp = dirfirst("*.c");
+	tmp = dirfirst("*.c", False);
 	printf("\tdirfirst(\"*.c\") = \"%s\"\n", tmp);
 	while ((tmp = dirnext()) != NULL)
 	{
@@ -230,7 +367,7 @@ BOOLEAN ostext(void)
 	}
 
 	/* test writing */
-	if (txtopen("verify.txt", 'w'))
+	if (txtopen("verify.txt", 'w', False))
 	{
 		printf("\ttxtopen(\"verify.txt\", 'w') failed\n");
 		return False;
@@ -244,7 +381,7 @@ BOOLEAN ostext(void)
 	txtclose();
 
 	/* readback, and compare */
-	if (txtopen("verify.txt", 'r'))
+	if (txtopen("verify.txt", 'r', False))
 	{
 		printf("\ttxtopen(\"verify.txt\", 'r') failed\n");
 		return False;
@@ -263,7 +400,7 @@ BOOLEAN ostext(void)
 	}
 
 	/* test appending */
-	if (txtopen("verify.txt", 'a'))
+	if (txtopen("verify.txt", 'a', False))
 	{
 		printf("\ttxtopen(\"verify.txt\", 'a') failed\n");
 		return False;
@@ -277,7 +414,7 @@ BOOLEAN ostext(void)
 	txtclose();
 
 	/* readback, and compare */
-	if (txtopen("verify.txt", 'r'))
+	if (txtopen("verify.txt", 'r', False))
 	{
 		printf("\ttxtopen(\"verify.txt\", 'r') failed\n");
 		return False;
@@ -303,7 +440,7 @@ BOOLEAN ostext(void)
 BOOLEAN options(void)
 {
 	BOOLEAN	passed = True;
-	char	outbuf[100];
+	CHAR	outbuf[100];
 	char	argbuf[20];
 
 	/* boolean testing */
@@ -334,11 +471,11 @@ BOOLEAN options(void)
 #endif
 
 	/* string testing */
-	if (!optset(False, strcpy(argbuf, "sh=csh"), outbuf, sizeof(outbuf))
+	if (!optset(False, (CHAR *)strcpy(argbuf, "sh=csh"), outbuf, sizeof(outbuf))
 	 || !o_shell
-	 || strcmp(o_shell, "csh")
-	 || !optset(False, strcpy(argbuf, "shell"), outbuf, sizeof(outbuf))
-	 || !strcmp(outbuf, "shell=sh\n")
+	 || strcmp(tochar8(o_shell), "csh")
+	 || !optset(False, (CHAR *)strcpy(argbuf, "shell"), outbuf, sizeof(outbuf))
+	 || !strcmp(tochar8(outbuf), "shell=sh\n")
 	 || (optflags(o_shell) & OPT_SET) == 0)
 	{
 		printf("\ttrouble with string options\n");
@@ -346,10 +483,10 @@ BOOLEAN options(void)
 	}
 
 	/* number testing */
-	if (!optset(False, strcpy(argbuf, "window=17"), outbuf, sizeof(outbuf))
+	if (!optset(False, (CHAR *)strcpy(argbuf, "window=17"), outbuf, sizeof(outbuf))
 	 || o_window != 17
-	 || !optset(False, strcpy(argbuf, "window"), outbuf, sizeof(outbuf))
-	 || strcmp(outbuf, "window=17\n")
+	 || !optset(False, (CHAR *)strcpy(argbuf, "window"), outbuf, sizeof(outbuf))
+	 || strcmp(tochar8(outbuf), "window=17\n")
 	 || (optflags(o_window) & OPT_SET) == 0)
 	{
 		printf("\ttrouble with number options\n");
@@ -384,7 +521,7 @@ BOOLEAN session(void)
 	sesunlock(0, True);
 
 	/* allocate another block */
-	blkno1 = sesalloc(0);
+	blkno1 = sesalloc(0, SES_CHARS);
 	if (blkno1 != 1)
 	{
 		printf("\tbad allocation of first block -- got %d (expect 1)\n", blkno1);
@@ -397,7 +534,7 @@ BOOLEAN session(void)
 		passed = False;
 	}
 	sesunlock(blkno1, False);
-	blkno2 = sesalloc(blkno1);
+	blkno2 = sesalloc(blkno1, SES_CHARS);
 	if (blkno2 != 1)
 	{
 		printf("\tsesalloc of block 1 failed -- got %d (expect 1)\n", blkno2);
@@ -481,32 +618,28 @@ BOOLEAN buffer(void)
 	MARKBUF	mark;
 	int	i;
 
-	o_session = "verify.elv";
-#if 1
+	o_session = toCHAR("verify.ses");
 	bufinit();
-#else
-	sesopen(False);
-#endif
-	buf = bufalloc("buffer", 0);
+	buf = bufalloc(toCHAR("buffer"), 0, False);
 	(void)marktmp(mark, buf, 0);
 	for (i = 0; i < 200; i++)
 	{
-		bufreplace(&mark, &mark, "This is a test\n", 15);
+		bufreplace(&mark, &mark, toCHAR("This is a test\n"), 15);
 	}
 	if (o_buflines(buf) != 200)
 	{
-		printf("o_buflines(buf) = %d (expect 200)\n", o_buflines(buf));
+		printf("o_buflines(buf) = %ld (expect 200)\n", o_buflines(buf));
 		passed = False;
 	}
 	if (o_bufchars(buf) != 3000)
 	{
-		printf("o_bufchars(buf) = %d (expect 3000)\n", o_bufchars(buf));
+		printf("o_bufchars(buf) = %ld (expect 3000)\n", o_bufchars(buf));
 		passed = False;
 	}
 	win = winalloc(NULL, NULL, buf, 24, 80);
 	while (markoffset(&mark) < o_bufchars(buf))
 	{
-		modehex.image(win, &mark, (ELVISSTATE)0, drawchar);
+		dmhex.image(win, &mark, (ELVISSTATE)0, drawchar);
 	}
 	sesclose();
 	return passed;
