@@ -15,10 +15,11 @@ typedef struct window_s
 	MARK	seltop;		/* start of selected text, or NULL */
 	MARK	selbottom;	/* end of selected text, or NULL */
 	long	selleft;	/* left column limit for marking (0 for any) */
-	long	selright;	/* right column limit (INIFINITY for any) */
+	long	selright;	/* right column limit (INFINITY for any) */
 	long	selorigcol;	/* column where cursor was when marking began */
 	BOOLEAN	selattop;	/* boolean: does seltop follow cursor? (else selbottom does) */
 	CHAR	seltype;	/* 'c'=character, 'l'=line, 'r'=rectangle */
+	CHAR	cmdchars[7];	/* up to six characters of a command */
 	long	match;		/* offset of matching parenthesis, for showmatch */
 	long	wantcol;	/* column where cursor wants to be */
 	int	cursx, cursy;	/* cursor position on window */
@@ -41,7 +42,9 @@ typedef struct window_s
 	OPTVAL	showmatch;	/* show word wrap? */
 	OPTVAL	showmode;	/* show command mode? */
 	OPTVAL	showstack;	/* show state stack? */
-	OPTVAL	sidescroll;	/* how to handle long lines */
+	OPTVAL	showcmd;	/* show command characters? */
+	OPTVAL	wrap;		/* wrap long lines? (else scroll sideways) */
+	OPTVAL	sidescroll;	/* scroll size when scrolling sideways */
 	OPTVAL	wrapmargin;	/* word wrap margin */
 } WINDOWBUF, *WINDOW;
 
@@ -56,9 +59,12 @@ typedef struct window_s
 #define o_showmatch(win)	((win)->showmatch.value.boolean)
 #define o_showmode(win)		((win)->showmode.value.boolean)
 #define o_showstack(win)	((win)->showstack.value.boolean)
+#define o_showcmd(win)		((win)->showcmd.value.boolean)
+#define o_wrap(win)		((win)->wrap.value.boolean)
 #define o_sidescroll(win)	((win)->sidescroll.value.number)
-#define o_wrapmargin(win)	use_o_linelength_instead
+#define o_wrapmargin(win)	use_o_textwidth_instead
 
+BEGIN_EXTERNC
 extern void wininit P_((void));
 extern WINDOW winalloc P_((GUIWIN *gw, OPTVAL *gvals, BUFFER buf, long rows, long columns));
 extern void winfree P_((WINDOW win, BOOLEAN force));
@@ -68,7 +74,7 @@ extern int wincount P_((BUFFER buf));
 extern void winchgbuf P_((WINDOW win, BUFFER newbuf, BOOLEAN force));
 extern WINDOW winofbuf P_((WINDOW win, BUFFER buf));
 extern WINDOW winofgw P_((GUIWIN *gw));
-extern WINDOW winofid P_((long id));
+END_EXTERNC
 
 extern WINDOW windows;
 extern WINDOW windefault;
